import os
import glob
import argparse
from PIL import Image
from torch.utils.data import Dataset, DataLoader
from torchvision import transforms

from subset_utils import motion_statistics, filtering_vis, get_relative_path


class DavisTest(Dataset):
    def __init__(self, img_size, data_root, split):
        nbr_frame = 2
        self.data_root = data_root

        self.transforms = transforms.Compose(
            [transforms.Resize((img_size, img_size)), transforms.ToTensor()]
        )
        if split == "trainval":
            all_folders = glob.glob(
                os.path.join(
                    self.data_root,
                    "davis2017/DAVIS-2017-trainval-480p",
                    "DAVIS/JPEGImages/480p",
                    "*",
                )
            )
        elif split == "test":
            all_folders = glob.glob(
                os.path.join(
                    self.data_root,
                    "davis2017/DAVIS-2017-test-dev-480p",
                    "DAVIS/JPEGImages/480p",
                    "*",
                )
            )

        print("original dataset has {} folders".format(len(all_folders)))
        folders = []
        num_samples = [0]
        total_num = 0

        # record the accumulated num of samples from the first video
        for f in all_folders:
            if len(glob.glob(os.path.join(f, "*"))) >= nbr_frame:
                folders.append(f)
                total_num += len(glob.glob(os.path.join(f, "*"))) // nbr_frame
                num_samples.append(total_num)

        self.folders = folders
        self.nbr_frame = nbr_frame
        self.num_samples = num_samples

    def __getitem__(self, index):
        # get frame pair sample given an index
        closest_value = min(self.num_samples, key=lambda x: abs(x - index))
        closest_index = self.num_samples.index(closest_value)
        if closest_value > index:
            folder = self.folders[closest_index - 1]
            start_frame = (index - self.num_samples[closest_index - 1]) * self.nbr_frame
        elif closest_value == index:
            folder = self.folders[closest_index]
            start_frame = 0
        else:
            folder = self.folders[closest_index]
            start_frame = (index - closest_value) * self.nbr_frame
        paths = sorted(glob.glob(os.path.join(folder, "*")))[
            start_frame : start_frame + self.nbr_frame
        ]
        images = []
        image_paths = []
        for path in paths:
            images.append(self.transforms(Image.open(path)))
            image_paths.append(get_relative_path(path, self.data_root))

        return images, image_paths

    def __len__(self):
        return self.num_samples[-1]


def get_loader(split, img_size, data_root, batch_size, shuffle, num_workers):
    dataset = DavisTest(img_size, data_root, split)
    return DataLoader(
        dataset,
        batch_size=batch_size,
        shuffle=shuffle,
        num_workers=num_workers,
        pin_memory=True,
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Process some inputs.")
    parser.add_argument("--data_root", type=str, help="root folder to save data")
    args = parser.parse_args()
    data_root = args.data_root

    # calculate statistics for frame pairs
    loader = get_loader(
        "trainval",
        img_size=512,
        data_root=data_root,
        batch_size=1,
        shuffle=False,
        num_workers=4,
    )
    motion_statistics(data_root, loader, "davis", "trainval")

    loader = get_loader(
        "test",
        img_size=512,
        data_root=data_root,
        batch_size=1,
        shuffle=False,
        num_workers=4,
    )
    motion_statistics(data_root, loader, "davis", "test")

    # visualize with gif of frame pairs given a set of thresholds
    threshold_dict = {"max_dist": 20, "80th": 2, "0.01st": 0.1}
    mse_threshold = 10

    flow_data_path = os.path.join(data_root, "flow_data/davis_trainval.json")
    filtering_vis(
        data_root, "davis", "trainval", flow_data_path, threshold_dict, mse_threshold
    )

    flow_data_path = os.path.join(data_root, "flow_data/davis_test.json")
    filtering_vis(
        data_root, "davis", "test", flow_data_path, threshold_dict, mse_threshold
    )



"""
Utilities for flow prediction, including:
    torch modules for pwcnet and raft flow
    image warping
    flow clipping
"""

import argparse
import math
import pathlib

import torch
import torch.nn as nn
import torch.nn.functional as F


class RAFT(nn.Module):
    def __init__(self, model="things", num_iters=5, dropout=0):
        super(RAFT, self).__init__()

        from neurovc.contrib.flowmag.flow_models.raft import raft

        if model == "things":
            model = "raft-things.pth"
        else:
            raise NotImplementedError

        # Get location of checkpoints
        raft_dir = pathlib.Path(__file__).parent.absolute() / "flow_models" / "raft"

        # Emulate arguments
        args = argparse.Namespace()
        args.model = raft_dir / model
        args.small = False
        args.mixed_precision = True
        args.alternate_corr = False
        args.dropout = dropout

        flowNet = nn.DataParallel(raft.RAFT(args))
        flowNet.load_state_dict(torch.load(args.model, map_location="cpu"))
        self.flowNet = flowNet.module.cpu()

        self.num_iters = num_iters

    def forward(self, im1, im2):
        """
        Input: images \in [0,1]
        """
        # Normalize to [0, 255]
        im1 = im1 * 255
        im2 = im2 * 255

        # Estimate flow
        flow_low, flow_up = self.flowNet(im1, im2, iters=self.num_iters, test_mode=True)

        return flow_up


class ARFlow(nn.Module):
    def __init__(self):
        super(ARFlow, self).__init__()

        from neurovc.contrib.flowmag.flow_models.ARFlow.models.pwclite import PWCLite
        from easydict import EasyDict
        from utils.torch_utils import restore_model

        chkpt_path = (
            pathlib.Path(__file__).parent.absolute()
            / "flow_models/ARFlow/checkpoints/KITTI15/pwclite_ar.tar"
        )

        model_cfg = {
            "upsample": True,
            "n_frames": 2,
            "reduce_dense": True,
        }
        model_cfg = EasyDict(model_cfg)
        flowNet = PWCLite(model_cfg)
        flowNet = restore_model(flowNet, chkpt_path)
        self.flowNet = flowNet

    def forward(self, im1, im2):
        """
        Input: images \in [0,1]
        """
        inp = torch.cat([im1, im2], dim=1)
        return self.flowNet(inp)["flows_fw"][0]


class GMFlow(nn.Module):
    def __init__(self, model="things"):
        super(GMFlow, self).__init__()

        from neurovc.contrib.flowmag.flow_models.gmflow import gmflow

        if model == "sintel":
            model = "gmflow_sintel-0c07dcb3.pth"
        elif model == "things":
            model = "gmflow_things-e9887eda.pth"
        else:
            raise NotImplementedError

        # Get location of checkpoints
        gmflow_dir = pathlib.Path(__file__).parent.absolute() / "flow_models" / "gmflow"
        chkpt_path = gmflow_dir / model

        flowNet = gmflow.GMFlow(
            feature_channels=128,
            num_scales=1,
            upsample_factor=8,
            num_head=1,
            attention_type="swin",
            ffn_dim_expansion=4,
            num_transformer_layers=6,
        )
        checkpoint = torch.load(chkpt_path, map_location="cpu")
        flowNet.load_state_dict(checkpoint["model"])
        self.flowNet = flowNet

    def forward(self, im1, im2):
        """
        Input: images \in [0,1]
        """
        # Normalize to [0, 255]
        im1 = im1 * 255
        im2 = im2 * 255

        # Estimate flow
        results_dict = self.flowNet(
            im1,
            im2,
            attn_splits_list=[2],
            corr_radius_list=[-1],
            prop_radius_list=[-1],
            pred_bidir_flow=False,
        )
        return results_dict["flow_preds"][-1]


class PWC(nn.Module):
    def __init__(self):
        super(PWC, self).__init__()

        from neurovc.contrib.flowmag.flow_models.pwcnet.pwc import Network

        self.flowNet = Network().eval().cpu()

    def forward(self, im1, im2):
        im1 = im1.squeeze()
        im2 = im2.squeeze()

        intWidth = im1.shape[2]
        intHeight = im1.shape[1]

        tenPreprocessedFirst = im1.cuda().view(1, 3, intHeight, intWidth)
        tenPreprocessedSecond = im2.cuda().view(1, 3, intHeight, intWidth)

        intPreprocessedWidth = int(math.floor(math.ceil(intWidth / 64.0) * 64.0))
        intPreprocessedHeight = int(math.floor(math.ceil(intHeight / 64.0) * 64.0))

        tenPreprocessedFirst = torch.nn.functional.interpolate(
            input=tenPreprocessedFirst,
            size=(intPreprocessedHeight, intPreprocessedWidth),
            mode="bilinear",
            align_corners=False,
        )
        tenPreprocessedSecond = torch.nn.functional.interpolate(
            input=tenPreprocessedSecond,
            size=(intPreprocessedHeight, intPreprocessedWidth),
            mode="bilinear",
            align_corners=False,
        )

        tenFlow = 20.0 * torch.nn.functional.interpolate(
            input=self.flowNet(tenPreprocessedFirst, tenPreprocessedSecond),
            size=(intHeight, intWidth),
            mode="bilinear",
            align_corners=False,
        )

        tenFlow[:, 0, :, :] *= float(intWidth) / float(intPreprocessedWidth)
        tenFlow[:, 1, :, :] *= float(intHeight) / float(intPreprocessedHeight)

        return tenFlow


def normalize_flow(flow):
    """
    Normalize pixel-offset (relative) flow to absolute [-1, 1] flow
    input :
        flow : tensor (b, 2, h, w)
    output :
        flow : tensor (b, h, w, 2) (for `F.grid_sample`)
    """
    _, _, h, w = flow.shape
    device = flow.device

    # Get base pixel coordinates (just "gaussian integers")
    base = torch.meshgrid(torch.arange(h), torch.arange(w))[::-1]
    base = torch.stack(base).float().to(device)

    # Convert to absolute coordinates
    flow = flow + base

    # Convert to [-1, 1] for grid_sample
    size = torch.tensor([w, h]).float().to(device)
    flow = -1 + 2.0 * flow / (-1 + size)[:, None, None]
    flow = flow.permute(0, 2, 3, 1)

    return flow


def warp(im, flow, padding_mode="reflection"):
    """
    requires absolute flow, normalized to [-1, 1]
        (see `normalize_flow` function)
    """

    warped = F.grid_sample(im, flow, padding_mode=padding_mode, align_corners=True)

    return warped


if __name__ == "__main__":
    flowNet = ARFlow()

    im1 = torch.randn(8, 3, 512, 512)
    im2 = torch.randn(8, 3, 512, 512)

    flow = flowNet(im1, im2)

    _ = 1



__author__ = "Philipp Flotho"
"""
multimodal_cam_calib
Copyright 2021 by Philipp Flotho, All rights reserved.
"""
import numpy as np  # noqa: E402
import cv2  # noqa: E402
import os  # noqa: E402


class VideoLooper:
    def __init__(self):
        self.stop = False

    def __call__(self, keystroke):
        keystroke &= 0xFF
        if keystroke == ord(" "):
            while True:
                keystroke = cv2.waitKey() & 0xFF
                if keystroke == 27:
                    self.stop = True
                    break
                if keystroke == ord(" "):
                    break
        return keystroke == 27 or self.stop


class CircularFrameBuffer:
    def __init__(self, size):
        self.size = size
        self.buffer = []
        self.counter = 0

    def append(self, frame):
        if len(self.buffer) < self.size:
            self.buffer.append(frame)
        else:
            self.buffer[self.counter] = frame
            self.counter = (self.counter + 1) % self.size

    def get_oldest(self):
        if len(self.buffer) == 0:
            return None
        return self.buffer[self.counter]

    def get_newest(self):
        if len(self.buffer) == 0:
            return None
        return self.buffer[(self.counter - 1) % self.size]

    def get(self):
        return self.buffer


class VideoReader:
    def __init__(self, filepath):
        self.cap = cv2.VideoCapture(filepath)
        self.framerate = self.cap.get(cv2.CAP_PROP_FPS)
        self.n_frames = self.cap.get(cv2.CAP_PROP_FRAME_COUNT)

    def read_frame(self, i=None):
        ret = False
        frame = None
        if self.cap.isOpened():
            if i is not None and i < self.n_frames:
                self.cap.set(cv2.CAP_PROP_POS_FRAMES, i)
            ret, frame = self.cap.read()
        return ret, frame

    def has_frames(self):
        return (
            self.cap.isOpened()
            and self.cap.get(cv2.CAP_PROP_POS_FRAMES) < self.n_frames
        )

    def read_frames(self, start_idx=0, end_idx=-1, offset=1):
        self.cap.set(cv2.CAP_PROP_POS_FRAMES, start_idx)
        counter = start_idx
        frames = []
        while counter < end_idx:
            ret, frame = self.cap.read()
            if not ret:
                break
            frames.append(frame)
            counter += 1
        return ret, np.array(frames)

    def __getitem__(self, item):
        if isinstance(item, slice):
            start, stop, step = item.indices(int(self.n_frames))
            success, frames = self.read_frames(
                start_idx=start, end_idx=stop, offset=step
            )
            return frames if success else np.array([])
        if isinstance(item, int):
            idx = item
            if idx < 0:
                idx = int(self.n_frames) + idx
            ret, frame = self.read_frame(idx)
            return frame if ret else None
        raise TypeError("VideoReader indices must be integers or slices.")


class HDFFileVideoWriter:
    def __init__(self, filepath="output.h5", ds_names=("Frames",)):
        try:
            from h5py import File as _File
        except ImportError as exc:
            raise ImportError(
                "h5py is required for HDFFileVideoWriter. "
                "Install via `pip install h5py` or `pip install neurovc[momag]`."
            ) from exc
        self._File = _File
        self.__init = False
        self.ds_names = ds_names
        try:
            os.remove(filepath)
        except OSError:
            pass
        self.file = self._File(filepath, "w")

    def __call__(self, frames, ts):
        assert len(frames) == len(self.ds_names)
        if not self.__init:
            for i, frame in enumerate(frames):
                self.file.create_dataset(
                    self.ds_names[i],
                    (0,) + frame.shape,
                    dtype=frame.dtype,
                    chunks=(1,) + frame.shape,
                    maxshape=(None,) + frame.shape,
                )
            self.file.create_dataset("Timestamps", (0,), chunks=True, maxshape=(None,))
            self.__init = True

        for i, frame in enumerate(frames):
            ds = self.file[self.ds_names[i]]
            ds.resize((ds.shape[0] + 1,) + ds.shape[1:])
            ds[-1] = frame
        ds = self.file["Timestamps"]
        ds.resize((ds.shape[0] + 1,))
        ds[-1] = ts

    def __del__(self):
        if self.file is not None:
            self.file.close()


class VideoWriter:
    def __init__(self, filepath, framerate=50):
        self.writer = None
        self.filepath = filepath
        self.width = None
        self.height = None
        self.n_channels = None
        self.framerate = framerate

    def _check_range(self, min_val, max_val, checkrange):
        return

    def __call__(self, frame):
        if self.writer is None:
            self.height, self.width = frame.shape[:2]

            fourcc = cv2.VideoWriter_fourcc(
                *"mp4v"
            )  # cv2.VideoWriter_fourcc('M', 'J', 'P', 'G')
            self.writer = cv2.VideoWriter(
                self.filepath, fourcc, self.framerate, (self.width, self.height)
            )
            # self.writer = cv2.VideoWriter(self.filepath, cv2.VideoWriter_fourcc('D', 'I', 'V', 'X'), self.framerate, (self.width, self.height))

        if len(frame.shape) < 3:
            self.n_channels = 1
        else:
            self.n_channels = frame.shape[2]

        if self.n_channels == 1:
            frame = cv2.cvtColor(frame, cv2.COLOR_GRAY2RGB)

        if frame.dtype != np.uint8:
            min_val = np.min(frame)
            max_val = np.max(frame)
            if min_val > 0 and max_val < 1:
                frame = (255 * frame).astype(np.uint8)
            elif min_val > 0 and max_val < 255:
                frame = frame.astype(np.uint8)
            else:
                frame = cv2.normalize(
                    frame,
                    None,
                    alpha=0,
                    beta=255,
                    norm_type=cv2.NORM_MINMAX,
                    dtype=cv2.CV_8UC3,
                )
        self.writer.write(frame)

    def write_frames(self, frames):
        for frame in frames:
            self(frame)

    def __del__(self):
        if self.writer is not None:
            self.writer.release()


class Debayerer:
    def __init__(self, pattern=cv2.COLOR_BAYER_RG2BGR, nh=2):
        self.pattern = pattern
        colors = np.array([255, 255, 255])

        self.reference = np.expand_dims(
            np.mean(np.array(colors), axis=0).astype(float), (0, 1)
        )

    def set_white_balance(self, colors):
        if isinstance(colors, np.ndarray):
            assert np.sum(colors.shape) == 3
        else:
            assert len(colors) == 3
        self.reference = np.expand_dims(
            np.mean(np.squeeze(np.array(colors)), axis=0).astype(float), (0, 1)
        )

    def __call__(self, img):
        tmp = (cv2.cvtColor(img, self.pattern).astype(float) / self.reference) * 255
        tmp[tmp > 255] = 255
        return tmp.astype(np.uint8)


def remap_RGB(frame, scaling, f1, f2):
    frame = scaling * (frame.astype(float) - f2) - f1
    frame[frame < 0] = 0
    frame[frame > 255] = 255
    return frame.astype(np.uint8)


def imagesc(img, window_name="Imagesc", color_map=cv2.COLORMAP_HOT):
    cv2.imshow(window_name, normalize_color(img, color_map))


def normalize_color(img, color_map=cv2.COLORMAP_HOT, normalize=True):
    img_8b = np.empty(img.shape, np.uint8)
    if normalize:
        cv2.normalize(
            img, img_8b, alpha=0, beta=255, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_8UC1
        )
    else:
        img_8b = (255 * img).astype(np.uint8)
    img_8b = cv2.applyColorMap(img_8b, color_map)
    return img_8b


def draw_landmarks(frame, landmarks, weights=None, color=[0, 0, 0]):
    if frame.shape[2] != 3:
        frame_rgb = normalize_color(frame)
    else:
        frame_rgb = frame.copy()
    if weights is not None:
        weights -= weights.min()
        weights /= weights.max()
    for i, (x, y) in enumerate(landmarks[..., :2]):
        if weights is not None:
            color = [0, 0, 0]
            color[2] = 1 - weights[i]
            color[1] = weights[i]
            color = tuple((255 * np.array(color)).astype(int).tolist())
        cv2.circle(frame_rgb, (int(np.round(x)), int(np.round(y))), 2, color, -1)

    return frame_rgb


def imgaussfilt(img, sigma):
    if np.isscalar(sigma):
        width = 2 * (np.ceil(6 * sigma) // 2) + 1
        height = width
    else:
        width = 2 * (np.ceil(6 * sigma[0]) // 2) + 1
        height = 2 * (np.ceil(6 * sigma[1]) // 2) + 1
    width = int(width)
    height = int(height)
    return cv2.GaussianBlur(img, (width, height), sigmaX=sigma, sigmaY=sigma)


def map_temp(data, cam="A655", resolution="high"):
    a = 1
    if resolution == "high":
        a = 0.1
    if cam == "A65":
        return (a * 0.4 * data) - 273.15
    elif cam == "A655":
        return (a * 0.1 * data) - 273.15
    else:
        print("camera not implemented!")



import torch
import shutil
import numpy as np
import random
import math
from torch.optim import Optimizer


def init_seed(seed):
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)
    random.seed(seed)


def weight_parameters(module):
    return [param for name, param in module.named_parameters() if "weight" in name]


def bias_parameters(module):
    return [param for name, param in module.named_parameters() if "bias" in name]


def load_checkpoint(model_path):
    weights = torch.load(model_path)
    epoch = None
    if "epoch" in weights:
        epoch = weights.pop("epoch")
    if "state_dict" in weights:
        state_dict = weights["state_dict"]
    else:
        state_dict = weights
    return epoch, state_dict


def save_checkpoint(save_path, states, file_prefixes, is_best, filename="ckpt.pth.tar"):
    def run_one_sample(save_path, state, prefix, is_best, filename):
        torch.save(state, save_path / "{}_{}".format(prefix, filename))
        if is_best:
            shutil.copyfile(
                save_path / "{}_{}".format(prefix, filename),
                save_path / "{}_model_best.pth.tar".format(prefix),
            )

    if not isinstance(file_prefixes, str):
        for prefix, state in zip(file_prefixes, states):
            run_one_sample(save_path, state, prefix, is_best, filename)

    else:
        run_one_sample(save_path, states, file_prefixes, is_best, filename)


def restore_model(model, pretrained_file):
    epoch, weights = load_checkpoint(pretrained_file)

    model_keys = set(model.state_dict().keys())
    weight_keys = set(weights.keys())

    # load weights by name
    weights_not_in_model = sorted(list(weight_keys - model_keys))
    model_not_in_weights = sorted(list(model_keys - weight_keys))
    if len(model_not_in_weights):
        print("Warning: There are weights in model but not in pre-trained.")
        for key in model_not_in_weights:
            print(key)
            weights[key] = model.state_dict()[key]
    if len(weights_not_in_model):
        print("Warning: There are pre-trained weights not in model.")
        for key in weights_not_in_model:
            print(key)
        from collections import OrderedDict

        new_weights = OrderedDict()
        for key in model_keys:
            new_weights[key] = weights[key]
        weights = new_weights

    model.load_state_dict(weights)
    return model


class AdamW(Optimizer):
    """Implements AdamW algorithm.

    It has been proposed in `Fixing Weight Decay Regularization in Adam`_.

    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): learning rate (default: 1e-3)
        betas (Tuple[float, float], optional): coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.999))
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-8)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)

    .. Fixing Weight Decay Regularization in Adam:
    https://arxiv.org/abs/1711.05101
    """

    def __init__(self, params, lr=1e-3, betas=(0.9, 0.999), eps=1e-8, weight_decay=0):
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay)
        super(AdamW, self).__init__(params, defaults)

    def step(self, closure=None):
        """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError(
                        "AdamW does not support sparse gradients, please consider SparseAdam instead"
                    )

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state["step"] = 0
                    # Exponential moving average of gradient values
                    state["exp_avg"] = torch.zeros_like(p.data)
                    # Exponential moving average of squared gradient values
                    state["exp_avg_sq"] = torch.zeros_like(p.data)

                exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]
                beta1, beta2 = group["betas"]

                state["step"] += 1

                # according to the paper, this penalty should come after the bias correction
                # if group['weight_decay'] != 0:
                #     grad = grad.add(group['weight_decay'], p.data)

                # Decay the first and second moment running average coefficient
                exp_avg.mul_(beta1).add_(1 - beta1, grad)
                exp_avg_sq.mul_(beta2).addcmul_(1 - beta2, grad, grad)

                denom = exp_avg_sq.sqrt().add_(group["eps"])

                bias_correction1 = 1 - beta1 ** state["step"]
                bias_correction2 = 1 - beta2 ** state["step"]
                step_size = group["lr"] * math.sqrt(bias_correction2) / bias_correction1

                p.data.addcdiv_(-step_size, exp_avg, denom)

                if group["weight_decay"] != 0:
                    p.data.add_(-group["weight_decay"], p.data)

        return loss



import os
import cv2
import sys
import tqdm
import json
import torch
import random
import shutil
import itertools
import numpy as np

sys.path.append(".")

from html_utils import HTML, save_gif
from neurovc.contrib.flowmag.flow_utils import RAFT


def get_relative_path(path, data_root):
    # remove root directory and get relative path
    return path.replace(os.path.abspath(data_root), "")[1:]


def get_absolute_path(path, data_root):
    # add root directory and get absolute path
    return os.path.join(data_root, path)


def motion_statistics(data_root, loader, dataset_name, split):
    # record the paths, mse, and percentiles of optical flow (motion) for frame pairs
    if not os.path.exists(os.path.join(data_root, "flow_data")):
        os.mkdir(os.path.join(data_root, "flow_data"))
    saved_path = os.path.join(
        data_root, "flow_data/{}_{}.json".format(dataset_name, split)
    )
    print(dataset_name, split, saved_path)
    model = RAFT(model="things", num_iters=20)
    model.cuda()
    model.eval()

    flow_data = {}
    with torch.no_grad():
        for i, (images, image_paths) in enumerate(tqdm.tqdm(loader)):
            flow = model(images[0].cuda(), images[1].cuda())
            flow = np.abs(flow.squeeze().detach().cpu().numpy())
            flow_norm = np.sqrt(flow[0] ** 2 + flow[1] ** 2)
            paths = list(itertools.chain.from_iterable(image_paths))
            imgs = [
                cv2.imread(get_absolute_path(path, data_root)).astype(float)
                for path in paths
            ]
            flow_data[i] = {
                "paths": tuple(sorted(paths)),
                "mse": ((imgs[0] - imgs[1]) ** 2).mean(),
                "max_dist": np.percentile(flow_norm, 99.9),
                "80th": np.percentile(flow_norm, 80),
                "0.01st": np.percentile(flow_norm, 0.01),
            }

    with open(saved_path, "w") as f:
        json.dump(flow_data, f, indent=4)


def filtering_vis(
    data_root, dataset_name, split, flow_data_path, threshold_dict, mse_threshold
):
    # filter frame pairs with thresholds, and visualize the gif of frame pairs in html for inspection
    original_folders = []

    vis_path = os.path.join(data_root, "{}_{}_vis".format(dataset_name, split))
    if os.path.exists(vis_path):
        shutil.rmtree(vis_path)
    os.makedirs(vis_path)

    with open(flow_data_path, "r") as f:
        flow_data = json.load(f)
    included_data = {}
    filtered_data = {}
    for i, (_, value) in enumerate(tqdm.tqdm(flow_data.items())):
        image_paths = value["paths"]
        if value["mse"] <= mse_threshold:
            save = False
        else:
            for key, threshold in threshold_dict.items():
                if value[key] > threshold:
                    save = False
                    break
                else:
                    save = True

        if not save:
            filtered_data[i] = value

        else:
            included_data[i] = value
            original_folders.append(image_paths[0].split("/")[-2])

    print(
        "number of the selected videos: {}/{}, from {} original folders".format(
            len(included_data), len(filtered_data), len(set(original_folders))
        )
    )

    with open(
        os.path.join(
            data_root, "flow_data/{}_{}_filtered.json".format(dataset_name, split)
        ),
        "w",
    ) as f:
        json.dump(filtered_data, f, indent=4)

    with open(
        os.path.join(
            data_root, "flow_data/{}_{}_included.json".format(dataset_name, split)
        ),
        "w",
    ) as f:
        json.dump(included_data, f, indent=4)

    print(
        "saving html to {}".format(
            os.path.join(data_root, "{}_{}_vis".format(dataset_name, split))
        )
    )
    html = HTML(
        os.path.join(data_root, "{}_{}_vis".format(dataset_name, split)), "subset"
    )
    html.add_header(data_root.split("/")[-1])

    num_vis = 30
    filtered_vis = random.sample(list(filtered_data.keys()), num_vis)
    included_vis = random.sample(list(included_data.keys()), num_vis)

    html.add_header(
        "dataset: {}, split: {}, threshold: {}, {} (MSE)".format(
            dataset_name, split, str(threshold_dict), mse_threshold
        )
    )
    html.add_header(
        "filtered videos, {}/{}".format(
            len(filtered_data), len(filtered_data) + len(included_data)
        )
    )
    ims = []
    txts = []
    links = []
    for n in filtered_vis:
        paths = [
            get_absolute_path(path, data_root) for path in filtered_data[n]["paths"]
        ]
        gif_path = save_gif(
            paths,
            os.path.join(data_root, "{}_{}_vis".format(dataset_name, split), "images"),
        )
        ims.append(gif_path)
        txts.append(str(filtered_data[n]))
        links.append(gif_path)
    html.add_images(ims, txts, links)

    html.add_header(
        "included videos, {}/{}".format(
            len(included_data), len(filtered_data) + len(included_data)
        )
    )
    ims = []
    txts = []
    links = []
    for n in included_vis:
        paths = [
            get_absolute_path(path, data_root) for path in included_data[n]["paths"]
        ]
        gif_path = save_gif(
            paths,
            os.path.join(data_root, "{}_{}_vis".format(dataset_name, split), "images"),
        )
        ims.append(gif_path)
        txts.append(str(included_data[n]))
        links.append(gif_path)
    html.add_images(ims, txts, links)
    html.save()



import argparse
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import torch
import torch.nn as nn
from omegaconf import OmegaConf
from torchvision.utils import save_image
from tqdm import tqdm

from neurovc.contrib.flowmag.dataset import FramesDataset, TrainingFramesDataset
from neurovc.contrib.flowmag.myutils import dist_transform, get_our_model, write_video
from neurovc.contrib.flowmag.test_time_adapt import test_time_adapt


def inference(
    model,
    frames_dataset,
    save_dir,
    alpha=2.0,
    max_alpha=16.0,
    mask=None,
    num_device=1,
    output_video=False,
):
    """
    Takes frames_dataset, which should be ordered frames from a video
        and magnifies motion with respect to the first frame
        and saves them to the log file
    """
    device = "cuda"
    save_dir.mkdir(exist_ok=True, parents=True)
    results = []

    if isinstance(model, nn.Module):
        model.eval()

    training_status = model.module.get_training_status()

    # If input alpha exceeds the range for training, perform recursions for inference
    if alpha > max_alpha and np.sqrt(alpha) < max_alpha:
        our_alpha = np.sqrt(alpha)
        num_recursion = 2
    elif alpha < max_alpha:
        our_alpha = alpha
        num_recursion = 1
    else:
        raise Exception("alpha out of range")

    with torch.no_grad():
        im0 = frames_dataset[0][None].to(device)
        results.append(im0.detach().cpu())

        for i in tqdm(range(1, len(frames_dataset))):
            # Get i^th frame, and merge with 0^th frame
            im1 = frames_dataset[i][None].to(device)
            frames = torch.stack([im0, im1], dim=2).repeat(num_device, 1, 1, 1, 1)

            # Process frames
            for _ in range(num_recursion):
                if training_status:
                    pred, _, _ = model(frames, alpha=our_alpha, mask=mask)
                else:
                    pred = model(frames, alpha=our_alpha, mask=mask)
                frames = torch.stack([im0, pred[0, :, 0].unsqueeze(0)], dim=2).repeat(
                    num_device, 1, 1, 1, 1
                )

            # Save predicted frame
            pred = pred[0, :, 0]
            results.append(pred.detach().cpu())

    # Save as video file
    if output_video:
        saved_frames = [
            (255 * img.squeeze().permute(1, 2, 0).flip([-1]).numpy()).astype(np.uint8)
            for img in results
        ]
        video_path = (
            str(save_dir / f"x{alpha}.mp4")
            if mask is None
            else str(save_dir / f"masked_x{alpha}.mp4")
        )
        write_video(saved_frames, 30, video_path)
        print("saved the video to {}".format(video_path))

    # Save as image files
    else:
        save_dir = (
            save_dir / f"x{alpha}" if mask is None else save_dir / f"masked_x{alpha}"
        )
        save_dir.mkdir(exist_ok=True, parents=True)
        for i, img in enumerate(results):
            save_image(img, save_dir / f"{i+1:04}.png")
        print("saved the images to {}".format(str(save_dir)))


if __name__ == "__main__":
    # Load args and config
    parser = argparse.ArgumentParser(description="Process some integers.")
    parser.add_argument("--config", type=str, help="path to config file")
    parser.add_argument(
        "--frames_dir",
        type=str,
        required=True,
        help="path to directory of frames to magnify",
    )
    parser.add_argument("--resume", type=str, help="path to checkpoint to resume from")
    parser.add_argument(
        "--save_name", type=str, required=True, help="name to save under"
    )
    parser.add_argument(
        "--alpha", type=float, required=True, help="amount to magnify motion"
    )
    parser.add_argument(
        "--mask_path",
        type=str,
        default=None,
        help="path to numpy mask, if None then no mask",
    )
    parser.add_argument(
        "--soft_mask",
        type=int,
        default=0,
        help="how much to soften mask. 0 is none, higher is more",
    )
    parser.add_argument("--output_video", action="store_true")
    parser.add_argument("--test_time_adapt", action="store_true")
    parser.add_argument(
        "--tta_epoch",
        type=int,
        default=3,
        help="number of epochs for test time adaptation",
    )
    args = parser.parse_args()

    # Make frames dataset
    frames_dataset = (
        TrainingFramesDataset(args.frames_dir)
        if args.test_time_adapt
        else FramesDataset(args.frames_dir)
    )

    # Load and preprocess mask, if given
    mask = None
    if args.mask_path:
        # Load mask from npy file
        mask = np.load(args.mask_path)
        mask = torch.tensor(mask)
        h, w = mask.shape
        mask = mask.float()

        if args.soft_mask:
            print("Softening mask")
            dist = dist_transform(mask)
            dist[dist < args.soft_mask] = 1
            dist[dist >= args.soft_mask] = 0
            mask = dist

    # Get max alpha
    config = OmegaConf.load(args.config)
    max_alpha = config.train.alpha_high

    # Make save dir
    save_dir = Path(args.resume).parent.parent / "inference" / args.save_name
    save_dir.mkdir(exist_ok=True, parents=True)

    # Make model
    model, epoch = get_our_model(args, args.test_time_adapt)

    # Test time adaptation
    if args.test_time_adapt:
        save_dir = save_dir / f"tta_epoch{epoch:03}"
        save_dir.mkdir(exist_ok=True, parents=True)

        def inference_fn(model, epoch):
            new_save_dir = save_dir / f"tta_epoch{epoch:03}"
            new_save_dir.mkdir(exist_ok=True, parents=True)
            inference(
                model,
                frames_dataset,
                new_save_dir,
                alpha=args.alpha,
                max_alpha=max_alpha,
                mask=mask,
                num_device=1,
                output_video=args.output_video,
            )

        # Run test time adaptation
        model, loss_info = test_time_adapt(
            model,
            args.frames_dir,
            num_epochs=args.tta_epoch,
            inference_fn=inference_fn,
            inference_freq=1,
            alpha=None,
            save_dir=save_dir,
            dataset_length=50000,
        )

        # Save loss curve as images
        for loss_name, losses in loss_info.items():
            plt.plot(losses)
            plt.title(loss_name)
            plt.savefig(save_dir / f"{loss_name}.png")
            plt.clf()

    inference(
        model,
        frames_dataset,
        save_dir,
        alpha=args.alpha,
        max_alpha=max_alpha,
        mask=mask,
        num_device=1,
        output_video=args.output_video,
    )



import torch
import cv2
import numpy as np
from matplotlib.colors import hsv_to_rgb


def load_flow(path):
    if path.endswith(".png"):
        # for KITTI which uses 16bit PNG images
        # see 'https://github.com/ClementPinard/FlowNetPytorch/blob/master/datasets/KITTI.py'
        # The -1 is here to specify not to change the image depth (16bit), and is compatible
        # with both OpenCV2 and OpenCV3
        flo_file = cv2.imread(path, -1)
        flo_img = flo_file[:, :, 2:0:-1].astype(np.float32)
        invalid = flo_file[:, :, 0] == 0  # mask
        flo_img = flo_img - 32768
        flo_img = flo_img / 64
        flo_img[np.abs(flo_img) < 1e-10] = 1e-10
        flo_img[invalid, :] = 0
        return flo_img, np.expand_dims(flo_file[:, :, 0], 2)
    else:
        with open(path, "rb") as f:
            magic = np.fromfile(f, np.float32, count=1)
            assert 202021.25 == magic, "Magic number incorrect. Invalid .flo file"
            h = np.fromfile(f, np.int32, count=1)[0]
            w = np.fromfile(f, np.int32, count=1)[0]
            data = np.fromfile(f, np.float32, count=2 * w * h)
        # Reshape data into 3D array (columns, rows, bands)
        data2D = np.resize(data, (w, h, 2))
        return data2D


def flow_to_image(flow, max_flow=256):
    if max_flow is not None:
        max_flow = max(max_flow, 1.0)
    else:
        max_flow = np.max(flow)

    n = 8
    u, v = flow[:, :, 0], flow[:, :, 1]
    mag = np.sqrt(np.square(u) + np.square(v))
    angle = np.arctan2(v, u)
    im_h = np.mod(angle / (2 * np.pi) + 1, 1)
    im_s = np.clip(mag * n / max_flow, a_min=0, a_max=1)
    im_v = np.clip(n - im_s, a_min=0, a_max=1)
    im = hsv_to_rgb(np.stack([im_h, im_s, im_v], 2))
    return (im * 255).astype(np.uint8)


def resize_flow(flow, new_shape):
    _, _, h, w = flow.shape
    new_h, new_w = new_shape
    flow = torch.nn.functional.interpolate(
        flow, (new_h, new_w), mode="bilinear", align_corners=True
    )
    scale_h, scale_w = h / float(new_h), w / float(new_w)
    flow[:, 0] /= scale_w
    flow[:, 1] /= scale_h
    return flow


def evaluate_flow(gt_flows, pred_flows, moving_masks=None):
    # credit "undepthflow/eval/evaluate_flow.py"
    def calculate_error_rate(epe_map, gt_flow, mask):
        bad_pixels = np.logical_and(
            epe_map * mask > 3,
            epe_map
            * mask
            / np.maximum(np.sqrt(np.sum(np.square(gt_flow), axis=2)), 1e-10)
            > 0.05,
        )
        return bad_pixels.sum() / mask.sum() * 100.0

    error, error_noc, error_occ, error_move, error_static, error_rate = (
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
    )
    error_move_rate, error_static_rate = 0.0, 0.0
    B = len(gt_flows)
    for gt_flow, pred_flow, i in zip(gt_flows, pred_flows, range(B)):
        H, W = gt_flow.shape[:2]

        h, w = pred_flow.shape[:2]
        pred_flow = np.copy(pred_flow)
        pred_flow[:, :, 0] = pred_flow[:, :, 0] / w * W
        pred_flow[:, :, 1] = pred_flow[:, :, 1] / h * H

        flo_pred = cv2.resize(pred_flow, (W, H), interpolation=cv2.INTER_LINEAR)

        epe_map = np.sqrt(
            np.sum(np.square(flo_pred[:, :, :2] - gt_flow[:, :, :2]), axis=2)
        )
        if gt_flow.shape[-1] == 2:
            error += np.mean(epe_map)

        elif gt_flow.shape[-1] == 4:
            error += np.sum(epe_map * gt_flow[:, :, 2]) / np.sum(gt_flow[:, :, 2])
            noc_mask = gt_flow[:, :, -1]
            error_noc += np.sum(epe_map * noc_mask) / np.sum(noc_mask)

            error_occ += np.sum(epe_map * (gt_flow[:, :, 2] - noc_mask)) / max(
                np.sum(gt_flow[:, :, 2] - noc_mask), 1.0
            )

            error_rate += calculate_error_rate(
                epe_map, gt_flow[:, :, 0:2], gt_flow[:, :, 2]
            )

            if moving_masks is not None:
                move_mask = moving_masks[i]

                error_move_rate += calculate_error_rate(
                    epe_map, gt_flow[:, :, 0:2], gt_flow[:, :, 2] * move_mask
                )
                error_static_rate += calculate_error_rate(
                    epe_map, gt_flow[:, :, 0:2], gt_flow[:, :, 2] * (1.0 - move_mask)
                )

                error_move += np.sum(epe_map * gt_flow[:, :, 2] * move_mask) / np.sum(
                    gt_flow[:, :, 2] * move_mask
                )
                error_static += np.sum(
                    epe_map * gt_flow[:, :, 2] * (1.0 - move_mask)
                ) / np.sum(gt_flow[:, :, 2] * (1.0 - move_mask))

    if gt_flows[0].shape[-1] == 4:
        res = [error / B, error_noc / B, error_occ / B, error_rate / B]
        if moving_masks is not None:
            res += [error_move / B, error_static / B]
        return res
    else:
        return [error / B]



from neurovc.contrib.flowmag.flow_models.ARFlow.trainer import (
    kitti_trainer,
    kitti_trainer_ar,
    sintel_trainer,
    sintel_trainer_ar,
)


def get_trainer(name):
    if name == "Sintel":
        TrainFramework = sintel_trainer.TrainFramework
    elif name == "Sintel_AR":
        TrainFramework = sintel_trainer_ar.TrainFramework
    elif name == "KITTI":
        TrainFramework = kitti_trainer.TrainFramework
    elif name == "KITTI_AR":
        TrainFramework = kitti_trainer_ar.TrainFramework
    else:
        raise NotImplementedError(name)

    return TrainFramework



# Flow visualization code used from https://github.com/tomrunia/OpticalFlow_Visualization


# MIT License
#
# Copyright (c) 2018 Tom Runia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to conditions.
#
# Author: Tom Runia
# Date Created: 2018-08-03

import numpy as np


def make_colorwheel():
    """
    Generates a color wheel for optical flow visualization as presented in:
        Baker et al. "A Database and Evaluation Methodology for Optical Flow" (ICCV, 2007)
        URL: http://vision.middlebury.edu/flow/flowEval-iccv07.pdf

    Code follows the original C++ source code of Daniel Scharstein.
    Code follows the the Matlab source code of Deqing Sun.

    Returns:
        np.ndarray: Color wheel
    """

    RY = 15
    YG = 6
    GC = 4
    CB = 11
    BM = 13
    MR = 6

    ncols = RY + YG + GC + CB + BM + MR
    colorwheel = np.zeros((ncols, 3))
    col = 0

    # RY
    colorwheel[0:RY, 0] = 255
    colorwheel[0:RY, 1] = np.floor(255 * np.arange(0, RY) / RY)
    col = col + RY
    # YG
    colorwheel[col : col + YG, 0] = 255 - np.floor(255 * np.arange(0, YG) / YG)
    colorwheel[col : col + YG, 1] = 255
    col = col + YG
    # GC
    colorwheel[col : col + GC, 1] = 255
    colorwheel[col : col + GC, 2] = np.floor(255 * np.arange(0, GC) / GC)
    col = col + GC
    # CB
    colorwheel[col : col + CB, 1] = 255 - np.floor(255 * np.arange(CB) / CB)
    colorwheel[col : col + CB, 2] = 255
    col = col + CB
    # BM
    colorwheel[col : col + BM, 2] = 255
    colorwheel[col : col + BM, 0] = np.floor(255 * np.arange(0, BM) / BM)
    col = col + BM
    # MR
    colorwheel[col : col + MR, 2] = 255 - np.floor(255 * np.arange(MR) / MR)
    colorwheel[col : col + MR, 0] = 255
    return colorwheel


def flow_uv_to_colors(u, v, convert_to_bgr=False):
    """
    Applies the flow color wheel to (possibly clipped) flow components u and v.

    According to the C++ source code of Daniel Scharstein
    According to the Matlab source code of Deqing Sun

    Args:
        u (np.ndarray): Input horizontal flow of shape [H,W]
        v (np.ndarray): Input vertical flow of shape [H,W]
        convert_to_bgr (bool, optional): Convert output image to BGR. Defaults to False.

    Returns:
        np.ndarray: Flow visualization image of shape [H,W,3]
    """
    flow_image = np.zeros((u.shape[0], u.shape[1], 3), np.uint8)
    colorwheel = make_colorwheel()  # shape [55x3]
    ncols = colorwheel.shape[0]
    rad = np.sqrt(np.square(u) + np.square(v))
    a = np.arctan2(-v, -u) / np.pi
    fk = (a + 1) / 2 * (ncols - 1)
    k0 = np.floor(fk).astype(np.int32)
    k1 = k0 + 1
    k1[k1 == ncols] = 0
    f = fk - k0
    for i in range(colorwheel.shape[1]):
        tmp = colorwheel[:, i]
        col0 = tmp[k0] / 255.0
        col1 = tmp[k1] / 255.0
        col = (1 - f) * col0 + f * col1
        idx = rad <= 1
        col[idx] = 1 - rad[idx] * (1 - col[idx])
        col[~idx] = col[~idx] * 0.75  # out of range
        # Note the 2-i => BGR instead of RGB
        ch_idx = 2 - i if convert_to_bgr else i
        flow_image[:, :, ch_idx] = np.floor(255 * col)
    return flow_image


def flow_to_image(flow_uv, clip_flow=None, convert_to_bgr=False, scaling=None):
    """
    Expects a two dimensional flow image of shape.

    Args:
        flow_uv (np.ndarray): Flow UV image of shape [H,W,2]
        clip_flow (float, optional): Clip maximum of flow values. Defaults to None.
        convert_to_bgr (bool, optional): Convert output image to BGR. Defaults to False.

    Returns:
        np.ndarray: Flow visualization image of shape [H,W,3]
    """
    assert flow_uv.ndim == 3, "input flow must have three dimensions"
    assert flow_uv.shape[2] == 2, "input flow must have shape [H,W,2]"
    if clip_flow is not None:
        flow_uv = np.clip(flow_uv, -clip_flow, clip_flow)
    u = flow_uv[:, :, 0]
    v = flow_uv[:, :, 1]
    rad = np.sqrt(np.square(u) + np.square(v))
    rad_max = np.max(rad)
    epsilon = 1e-5
    if scaling is None:
        u = u / (rad_max + epsilon)
        v = v / (rad_max + epsilon)
    else:
        u = u / scaling
        v = v / scaling
    return flow_uv_to_colors(u, v, convert_to_bgr)



import torch.nn as nn
from neurovc.contrib.flowmag.models.unet_parts import (
    DoubleConv,
    Down,
    OutConv,
    Up,
)


ACTIVATION_MAP = {"relu": nn.ReLU, "tanh": nn.Tanh, "sigmoid": nn.Sigmoid}


class MLP(nn.Module):
    def __init__(self, num_neurons, activation="relu", final_activation=None):
        """
        num_neurons : list of num neurons per layer
        activation : one of ['relu', 'tanh', 'sigmoid']
        final_activation : same as `activation`, if None, then don't apply final activation
        """
        assert len(num_neurons) > 1, "Need more than one layer"
        assert activation in ACTIVATION_MAP, "Invalid activation!"

        super(MLP, self).__init__()

        self.layers = nn.ModuleList()

        for i in range(len(num_neurons) - 1):
            self.layers.append(nn.Linear(num_neurons[i], num_neurons[i + 1]))

        self.activation = ACTIVATION_MAP[activation]()

        if final_activation is not None:
            self.final_activation = ACTIVATION_MAP[final_activation]()
        else:
            self.final_activation = None

    def forward(self, x):
        for layer_idx, layer in enumerate(self.layers):
            x = layer(x)

            # Don't apply activation after last layer
            if layer_idx != len(self.layers) - 1:
                x = self.activation(x)

        if self.final_activation is not None:
            x = self.final_activation(x)

        return x


class UNet(nn.Module):
    """
    Modified From: https://github.com/milesial/Pytorch-UNet

    n_channels - # input channels
    n_classes - # output channels
    num_layers - # downsampling layers == # upsampling layers
                 (only downsamples num_layers - 1 times though
                 b/c one layer doesn't downsample)
    ndf - # of filters
    bilinear - If true use bilinear upsampling, else use transposed convs
    skip - if false, zero out skip connections
    return_bottleneck - if True, also return bottleneck
    """

    def __init__(
        self,
        n_channels=3,
        n_classes=3,
        num_layers=5,
        ndf=64,
        bilinear=True,
        skip=True,
        return_bottleneck=False,
        final_activation=None,
    ):
        super(UNet, self).__init__()
        self.n_channels = n_channels
        self.n_classes = n_classes
        self.bilinear = bilinear
        self.num_layers = num_layers
        self.ndf = ndf
        self.skip = skip
        self.return_bottleneck = return_bottleneck
        if final_activation is not None:
            self.final_activation = ACTIVATION_MAP[final_activation]()
        else:
            self.final_activation = None
        bilinear_factor = 2 if bilinear else 1

        self.down_layers = nn.ModuleList()
        for layer_idx in range(self.num_layers):
            mult = 2**layer_idx

            # First layer we just conv
            if layer_idx == 0:
                self.down_layers.append(DoubleConv(n_channels, ndf * mult))
            else:
                in_channels = ndf * mult // 2
                out_channels = ndf * mult

                # If last down layer
                if layer_idx == self.num_layers - 1:
                    out_channels = out_channels // bilinear_factor

                self.down_layers.append(Down(in_channels, out_channels))

        self.up_layers = nn.ModuleList()
        for layer_idx in range(self.num_layers):
            mult = 2 ** (self.num_layers - layer_idx - 1)

            in_channels = ndf * mult
            out_channels = ndf * mult // 2 // bilinear_factor

            # Last layer we just conv
            if layer_idx == self.num_layers - 1:
                self.up_layers.append(OutConv(in_channels, n_classes))

            # Second to last layer doesn't have bilinear factor
            elif layer_idx == self.num_layers - 2:
                out_channels = out_channels * bilinear_factor
                self.up_layers.append(Up(in_channels, out_channels, bilinear))

            else:
                self.up_layers.append(Up(in_channels, out_channels, bilinear))

    def encode(self, x):
        # Keep track of skip connections
        skips = []
        for layer in self.down_layers:
            x = layer(x)
            skips.append(x)

        bottleneck = skips.pop()

        return bottleneck, skips

    def decode(self, x, skips):
        # We want to access to skips in reverse order, ignoring the latest one
        skips = skips[::-1]
        # Append a dummy entry to match length of `up_layers` for `zip`
        skips.append(None)

        for idx, (layer, skip) in enumerate(zip(self.up_layers, skips)):
            if idx != self.num_layers - 1:
                if not self.skip:
                    skip = skip * 0
                x = layer(x, skip)

            # If last layer, no skip
            else:
                logits = layer(x)

        if self.final_activation is not None:
            logits = self.final_activation(logits)

        return logits

    def forward(self, x):
        bottleneck, skips = self.encode(x)

        logits = self.decode(bottleneck, skips)

        if self.return_bottleneck:
            return logits, bottleneck
        else:
            return logits


"""
model = MLP([784, 512, 256, 128, 64, 32, 10])
summary(model, (4,784))
"""

"""
model = UNet(3,3, ndf=16, num_layers=6, skip=False, return_bottleneck=True)
summary(model, (4,3,256,256))
"""



## Portions of Code from, copyright 2018 Jochen Gast

from __future__ import absolute_import, division, print_function

import torch
from torch import nn
import torch.nn.functional as tf


def _bchw2bhwc(tensor):
    return tensor.transpose(1, 2).transpose(2, 3)


def _bhwc2bchw(tensor):
    return tensor.transpose(2, 3).transpose(1, 2)


class Meshgrid(nn.Module):
    def __init__(self):
        super(Meshgrid, self).__init__()
        self.width = 0
        self.height = 0
        self.register_buffer("xx", torch.zeros(1, 1))
        self.register_buffer("yy", torch.zeros(1, 1))
        self.register_buffer("rangex", torch.zeros(1, 1))
        self.register_buffer("rangey", torch.zeros(1, 1))

    def _compute_meshgrid(self, width, height):
        torch.arange(0, width, out=self.rangex)
        torch.arange(0, height, out=self.rangey)
        self.xx = self.rangex.repeat(height, 1).contiguous()
        self.yy = self.rangey.repeat(width, 1).t().contiguous()

    def forward(self, width, height):
        if self.width != width or self.height != height:
            self._compute_meshgrid(width=width, height=height)
            self.width = width
            self.height = height
        return self.xx, self.yy


class BatchSub2Ind(nn.Module):
    def __init__(self):
        super(BatchSub2Ind, self).__init__()
        self.register_buffer("_offsets", torch.LongTensor())

    def forward(self, shape, row_sub, col_sub, out=None):
        batch_size = row_sub.size(0)
        height, width = shape
        ind = row_sub * width + col_sub
        torch.arange(batch_size, out=self._offsets)
        self._offsets *= height * width

        if out is None:
            return torch.add(ind, self._offsets.view(-1, 1, 1))
        else:
            torch.add(ind, self._offsets.view(-1, 1, 1), out=out)


class Interp2(nn.Module):
    def __init__(self, clamp=False):
        super(Interp2, self).__init__()
        self._clamp = clamp
        self._batch_sub2ind = BatchSub2Ind()
        self.register_buffer("_x0", torch.LongTensor())
        self.register_buffer("_x1", torch.LongTensor())
        self.register_buffer("_y0", torch.LongTensor())
        self.register_buffer("_y1", torch.LongTensor())
        self.register_buffer("_i00", torch.LongTensor())
        self.register_buffer("_i01", torch.LongTensor())
        self.register_buffer("_i10", torch.LongTensor())
        self.register_buffer("_i11", torch.LongTensor())
        self.register_buffer("_v00", torch.FloatTensor())
        self.register_buffer("_v01", torch.FloatTensor())
        self.register_buffer("_v10", torch.FloatTensor())
        self.register_buffer("_v11", torch.FloatTensor())
        self.register_buffer("_x", torch.FloatTensor())
        self.register_buffer("_y", torch.FloatTensor())

    def forward(self, v, xq, yq):
        batch_size, channels, height, width = v.size()

        # clamp if wanted
        if self._clamp:
            xq.clamp_(0, width - 1)
            yq.clamp_(0, height - 1)

        # ------------------------------------------------------------------
        # Find neighbors
        #
        # x0 = torch.floor(xq).long(),          x0.clamp_(0, width - 1)
        # x1 = x0 + 1,                          x1.clamp_(0, width - 1)
        # y0 = torch.floor(yq).long(),          y0.clamp_(0, height - 1)
        # y1 = y0 + 1,                          y1.clamp_(0, height - 1)
        #
        # ------------------------------------------------------------------
        self._x0 = torch.floor(xq).long().clamp(0, width - 1)
        self._y0 = torch.floor(yq).long().clamp(0, height - 1)

        self._x1 = torch.add(self._x0, 1).clamp(0, width - 1)
        self._y1 = torch.add(self._y0, 1).clamp(0, height - 1)

        # batch_sub2ind
        self._batch_sub2ind([height, width], self._y0, self._x0, out=self._i00)
        self._batch_sub2ind([height, width], self._y0, self._x1, out=self._i01)
        self._batch_sub2ind([height, width], self._y1, self._x0, out=self._i10)
        self._batch_sub2ind([height, width], self._y1, self._x1, out=self._i11)

        # reshape
        v_flat = _bchw2bhwc(v).contiguous().view(-1, channels)
        torch.index_select(v_flat, dim=0, index=self._i00.view(-1), out=self._v00)
        torch.index_select(v_flat, dim=0, index=self._i01.view(-1), out=self._v01)
        torch.index_select(v_flat, dim=0, index=self._i10.view(-1), out=self._v10)
        torch.index_select(v_flat, dim=0, index=self._i11.view(-1), out=self._v11)

        # local_coords
        torch.add(xq, -self._x0.float(), out=self._x)
        torch.add(yq, -self._y0.float(), out=self._y)

        # weights
        w00 = torch.unsqueeze((1.0 - self._y) * (1.0 - self._x), dim=1)
        w01 = torch.unsqueeze((1.0 - self._y) * self._x, dim=1)
        w10 = torch.unsqueeze(self._y * (1.0 - self._x), dim=1)
        w11 = torch.unsqueeze(self._y * self._x, dim=1)

        def _reshape(u):
            return _bhwc2bchw(u.view(batch_size, height, width, channels))

        # values
        values = (
            _reshape(self._v00) * w00
            + _reshape(self._v01) * w01
            + _reshape(self._v10) * w10
            + _reshape(self._v11) * w11
        )

        if self._clamp:
            return values
        else:
            #  find_invalid
            invalid = (
                ((xq < 0) | (xq >= width) | (yq < 0) | (yq >= height))
                .unsqueeze(dim=1)
                .float()
            )
            # maskout invalid
            transformed = invalid * torch.zeros_like(values) + (1.0 - invalid) * values

        return transformed


def resize2D(inputs, size_targets, mode="bilinear"):
    size_inputs = [inputs.size(2), inputs.size(3)]

    if all([size_inputs == size_targets]):
        return inputs  # nothing to do
    elif any([size_targets < size_inputs]):
        resized = tf.adaptive_avg_pool2d(inputs, size_targets)  # downscaling
    else:
        resized = tf.upsample(inputs, size=size_targets, mode=mode)  # upsampling

    # correct scaling
    return resized


def resize2D_as(inputs, output_as, mode="bilinear"):
    size_targets = [output_as.size(2), output_as.size(3)]
    return resize2D(inputs, size_targets, mode=mode)



import torch
import torch.nn.functional as F
from raft_utils.utils import bilinear_sampler

try:
    import alt_cuda_corr
except ImportError:
    # alt_cuda_corr is not compiled
    pass


class CorrBlock:
    def __init__(self, fmap1, fmap2, num_levels=4, radius=4):
        self.num_levels = num_levels
        self.radius = radius
        self.corr_pyramid = []

        # all pairs correlation
        corr = CorrBlock.corr(fmap1, fmap2)

        batch, h1, w1, dim, h2, w2 = corr.shape
        corr = corr.reshape(batch * h1 * w1, dim, h2, w2)

        self.corr_pyramid.append(corr)
        for i in range(self.num_levels - 1):
            corr = F.avg_pool2d(corr, 2, stride=2)
            self.corr_pyramid.append(corr)

    def __call__(self, coords):
        r = self.radius
        coords = coords.permute(0, 2, 3, 1)
        batch, h1, w1, _ = coords.shape

        out_pyramid = []
        for i in range(self.num_levels):
            corr = self.corr_pyramid[i]
            dx = torch.linspace(-r, r, 2 * r + 1)
            dy = torch.linspace(-r, r, 2 * r + 1)
            delta = torch.stack(torch.meshgrid(dy, dx), axis=-1).to(coords.device)

            centroid_lvl = coords.reshape(batch * h1 * w1, 1, 1, 2) / 2**i
            delta_lvl = delta.view(1, 2 * r + 1, 2 * r + 1, 2)
            coords_lvl = centroid_lvl + delta_lvl

            corr = bilinear_sampler(corr, coords_lvl)
            corr = corr.view(batch, h1, w1, -1)
            out_pyramid.append(corr)

        out = torch.cat(out_pyramid, dim=-1)
        return out.permute(0, 3, 1, 2).contiguous().float()

    @staticmethod
    def corr(fmap1, fmap2):
        batch, dim, ht, wd = fmap1.shape
        fmap1 = fmap1.view(batch, dim, ht * wd)
        fmap2 = fmap2.view(batch, dim, ht * wd)

        corr = torch.matmul(fmap1.transpose(1, 2), fmap2)
        corr = corr.view(batch, ht, wd, 1, ht, wd)
        return corr / torch.sqrt(torch.tensor(dim).float())


class AlternateCorrBlock:
    def __init__(self, fmap1, fmap2, num_levels=4, radius=4):
        self.num_levels = num_levels
        self.radius = radius

        self.pyramid = [(fmap1, fmap2)]
        for i in range(self.num_levels):
            fmap1 = F.avg_pool2d(fmap1, 2, stride=2)
            fmap2 = F.avg_pool2d(fmap2, 2, stride=2)
            self.pyramid.append((fmap1, fmap2))

    def __call__(self, coords):
        coords = coords.permute(0, 2, 3, 1)
        B, H, W, _ = coords.shape
        dim = self.pyramid[0][0].shape[1]

        corr_list = []
        for i in range(self.num_levels):
            r = self.radius
            fmap1_i = self.pyramid[0][0].permute(0, 2, 3, 1).contiguous()
            fmap2_i = self.pyramid[i][1].permute(0, 2, 3, 1).contiguous()

            coords_i = (coords / 2**i).reshape(B, 1, H, W, 2).contiguous()
            (corr,) = alt_cuda_corr.forward(fmap1_i, fmap2_i, coords_i, r)
            corr_list.append(corr.squeeze(1))

        corr = torch.stack(corr_list, dim=1)
        corr = corr.reshape(B, -1, H, W)
        return corr / torch.sqrt(torch.tensor(dim).float())



#!/usr/bin/env python3

from setuptools import setup
from torch.utils.cpp_extension import BuildExtension, CUDAExtension

cxx_args = ["-std=c++11"]

nvcc_args = [
    "-gencode",
    "arch=compute_50,code=sm_50",
    "-gencode",
    "arch=compute_52,code=sm_52",
    "-gencode",
    "arch=compute_60,code=sm_60",
    "-gencode",
    "arch=compute_61,code=sm_61",
    "-gencode",
    "arch=compute_61,code=compute_61",
    "-ccbin",
    "/usr/bin/gcc",
]

setup(
    name="correlation_cuda",
    ext_modules=[
        CUDAExtension(
            "correlation_cuda",
            ["correlation_cuda.cc", "correlation_cuda_kernel.cu"],
            extra_compile_args={
                "cxx": cxx_args,
                "nvcc": nvcc_args,
                "cuda-path": ["/usr/local/cuda-9.0"],
            },
        )
    ],
    cmdclass={"build_ext": BuildExtension},
)



#!/usr/bin/env python

import torch

import cupy
import re

kernel_Correlation_rearrange = """
	extern "C" __global__ void kernel_Correlation_rearrange(
		const int n,
		const float* input,
		float* output
	) {
	  int intIndex = (blockIdx.x * blockDim.x) + threadIdx.x;

	  if (intIndex >= n) {
	    return;
	  }

	  int intSample = blockIdx.z;
	  int intChannel = blockIdx.y;

	  float fltValue = input[(((intSample * SIZE_1(input)) + intChannel) * SIZE_2(input) * SIZE_3(input)) + intIndex];

	  __syncthreads();

	  int intPaddedY = (intIndex / SIZE_3(input)) + 4;
	  int intPaddedX = (intIndex % SIZE_3(input)) + 4;
	  int intRearrange = ((SIZE_3(input) + 8) * intPaddedY) + intPaddedX;

	  output[(((intSample * SIZE_1(output) * SIZE_2(output)) + intRearrange) * SIZE_1(input)) + intChannel] = fltValue;
	}
"""

kernel_Correlation_updateOutput = """
	extern "C" __global__ void kernel_Correlation_updateOutput(
	  const int n,
	  const float* rbot0,
	  const float* rbot1,
	  float* top
	) {
	  extern __shared__ char patch_data_char[];

	  float *patch_data = (float *)patch_data_char;

	  // First (upper left) position of kernel upper-left corner in current center position of neighborhood in image 1
	  int x1 = blockIdx.x + 4;
	  int y1 = blockIdx.y + 4;
	  int item = blockIdx.z;
	  int ch_off = threadIdx.x;

	  // Load 3D patch into shared shared memory
	  for (int j = 0; j < 1; j++) { // HEIGHT
	    for (int i = 0; i < 1; i++) { // WIDTH
	      int ji_off = (j + i) * SIZE_3(rbot0);
	      for (int ch = ch_off; ch < SIZE_3(rbot0); ch += 32) { // CHANNELS
	        int idx1 = ((item * SIZE_1(rbot0) + y1+j) * SIZE_2(rbot0) + x1+i) * SIZE_3(rbot0) + ch;
	        int idxPatchData = ji_off + ch;
	        patch_data[idxPatchData] = rbot0[idx1];
	      }
	    }
	  }

	  __syncthreads();

	  __shared__ float sum[32];

	  // Compute correlation
	  for (int top_channel = 0; top_channel < SIZE_1(top); top_channel++) {
	    sum[ch_off] = 0;

	    int s2o = top_channel % 9 - 4;
	    int s2p = top_channel / 9 - 4;

	    for (int j = 0; j < 1; j++) { // HEIGHT
	      for (int i = 0; i < 1; i++) { // WIDTH
	        int ji_off = (j + i) * SIZE_3(rbot0);
	        for (int ch = ch_off; ch < SIZE_3(rbot0); ch += 32) { // CHANNELS
	          int x2 = x1 + s2o;
	          int y2 = y1 + s2p;

	          int idxPatchData = ji_off + ch;
	          int idx2 = ((item * SIZE_1(rbot0) + y2+j) * SIZE_2(rbot0) + x2+i) * SIZE_3(rbot0) + ch;

	          sum[ch_off] += patch_data[idxPatchData] * rbot1[idx2];
	        }
	      }
	    }

	    __syncthreads();

	    if (ch_off == 0) {
	      float total_sum = 0;
	      for (int idx = 0; idx < 32; idx++) {
	        total_sum += sum[idx];
	      }
	      const int sumelems = SIZE_3(rbot0);
	      const int index = ((top_channel*SIZE_2(top) + blockIdx.y)*SIZE_3(top))+blockIdx.x;
	      top[index + item*SIZE_1(top)*SIZE_2(top)*SIZE_3(top)] = total_sum / (float)sumelems;
	    }
	  }
	}
"""

kernel_Correlation_updateGradFirst = """
	#define ROUND_OFF 50000

	extern "C" __global__ void kernel_Correlation_updateGradFirst(
	  const int n,
	  const int intSample,
	  const float* rbot0,
	  const float* rbot1,
	  const float* gradOutput,
	  float* gradFirst,
	  float* gradSecond
	) { for (int intIndex = (blockIdx.x * blockDim.x) + threadIdx.x; intIndex < n; intIndex += blockDim.x * gridDim.x) {
	  int n = intIndex % SIZE_1(gradFirst); // channels
	  int l = (intIndex / SIZE_1(gradFirst)) % SIZE_3(gradFirst) + 4; // w-pos
	  int m = (intIndex / SIZE_1(gradFirst) / SIZE_3(gradFirst)) % SIZE_2(gradFirst) + 4; // h-pos

	  // round_off is a trick to enable integer division with ceil, even for negative numbers
	  // We use a large offset, for the inner part not to become negative.
	  const int round_off = ROUND_OFF;
	  const int round_off_s1 = round_off;

	  // We add round_off before_s1 the int division and subtract round_off after it, to ensure the formula matches ceil behavior:
	  int xmin = (l - 4 + round_off_s1 - 1) + 1 - round_off; // ceil (l - 4)
	  int ymin = (m - 4 + round_off_s1 - 1) + 1 - round_off; // ceil (l - 4)

	  // Same here:
	  int xmax = (l - 4 + round_off_s1) - round_off; // floor (l - 4)
	  int ymax = (m - 4 + round_off_s1) - round_off; // floor (m - 4)

	  float sum = 0;
	  if (xmax>=0 && ymax>=0 && (xmin<=SIZE_3(gradOutput)-1) && (ymin<=SIZE_2(gradOutput)-1)) {
	    xmin = max(0,xmin);
	    xmax = min(SIZE_3(gradOutput)-1,xmax);

	    ymin = max(0,ymin);
	    ymax = min(SIZE_2(gradOutput)-1,ymax);

	    for (int p = -4; p <= 4; p++) {
	      for (int o = -4; o <= 4; o++) {
	        // Get rbot1 data:
	        int s2o = o;
	        int s2p = p;
	        int idxbot1 = ((intSample * SIZE_1(rbot0) + (m+s2p)) * SIZE_2(rbot0) + (l+s2o)) * SIZE_3(rbot0) + n;
	        float bot1tmp = rbot1[idxbot1]; // rbot1[l+s2o,m+s2p,n]

	        // Index offset for gradOutput in following loops:
	        int op = (p+4) * 9 + (o+4); // index[o,p]
	        int idxopoffset = (intSample * SIZE_1(gradOutput) + op);

	        for (int y = ymin; y <= ymax; y++) {
	          for (int x = xmin; x <= xmax; x++) {
	            int idxgradOutput = (idxopoffset * SIZE_2(gradOutput) + y) * SIZE_3(gradOutput) + x; // gradOutput[x,y,o,p]
	            sum += gradOutput[idxgradOutput] * bot1tmp;
	          }
	        }
	      }
	    }
	  }
	  const int sumelems = SIZE_1(gradFirst);
	  const int bot0index = ((n * SIZE_2(gradFirst)) + (m-4)) * SIZE_3(gradFirst) + (l-4);
	  gradFirst[bot0index + intSample*SIZE_1(gradFirst)*SIZE_2(gradFirst)*SIZE_3(gradFirst)] = sum / (float)sumelems;
	} }
"""

kernel_Correlation_updateGradSecond = """
	#define ROUND_OFF 50000

	extern "C" __global__ void kernel_Correlation_updateGradSecond(
	  const int n,
	  const int intSample,
	  const float* rbot0,
	  const float* rbot1,
	  const float* gradOutput,
	  float* gradFirst,
	  float* gradSecond
	) { for (int intIndex = (blockIdx.x * blockDim.x) + threadIdx.x; intIndex < n; intIndex += blockDim.x * gridDim.x) {
	  int n = intIndex % SIZE_1(gradSecond); // channels
	  int l = (intIndex / SIZE_1(gradSecond)) % SIZE_3(gradSecond) + 4; // w-pos
	  int m = (intIndex / SIZE_1(gradSecond) / SIZE_3(gradSecond)) % SIZE_2(gradSecond) + 4; // h-pos

	  // round_off is a trick to enable integer division with ceil, even for negative numbers
	  // We use a large offset, for the inner part not to become negative.
	  const int round_off = ROUND_OFF;
	  const int round_off_s1 = round_off;

	  float sum = 0;
	  for (int p = -4; p <= 4; p++) {
	    for (int o = -4; o <= 4; o++) {
	      int s2o = o;
	      int s2p = p;

	      //Get X,Y ranges and clamp
	      // We add round_off before_s1 the int division and subtract round_off after it, to ensure the formula matches ceil behavior:
	      int xmin = (l - 4 - s2o + round_off_s1 - 1) + 1 - round_off; // ceil (l - 4 - s2o)
	      int ymin = (m - 4 - s2p + round_off_s1 - 1) + 1 - round_off; // ceil (l - 4 - s2o)

	      // Same here:
	      int xmax = (l - 4 - s2o + round_off_s1) - round_off; // floor (l - 4 - s2o)
	      int ymax = (m - 4 - s2p + round_off_s1) - round_off; // floor (m - 4 - s2p)

	      if (xmax>=0 && ymax>=0 && (xmin<=SIZE_3(gradOutput)-1) && (ymin<=SIZE_2(gradOutput)-1)) {
	        xmin = max(0,xmin);
	        xmax = min(SIZE_3(gradOutput)-1,xmax);

	        ymin = max(0,ymin);
	        ymax = min(SIZE_2(gradOutput)-1,ymax);

	        // Get rbot0 data:
	        int idxbot0 = ((intSample * SIZE_1(rbot0) + (m-s2p)) * SIZE_2(rbot0) + (l-s2o)) * SIZE_3(rbot0) + n;
	        float bot0tmp = rbot0[idxbot0]; // rbot1[l+s2o,m+s2p,n]

	        // Index offset for gradOutput in following loops:
	        int op = (p+4) * 9 + (o+4); // index[o,p]
	        int idxopoffset = (intSample * SIZE_1(gradOutput) + op);

	        for (int y = ymin; y <= ymax; y++) {
	          for (int x = xmin; x <= xmax; x++) {
	            int idxgradOutput = (idxopoffset * SIZE_2(gradOutput) + y) * SIZE_3(gradOutput) + x; // gradOutput[x,y,o,p]
	            sum += gradOutput[idxgradOutput] * bot0tmp;
	          }
	        }
	      }
	    }
	  }
	  const int sumelems = SIZE_1(gradSecond);
	  const int bot1index = ((n * SIZE_2(gradSecond)) + (m-4)) * SIZE_3(gradSecond) + (l-4);
	  gradSecond[bot1index + intSample*SIZE_1(gradSecond)*SIZE_2(gradSecond)*SIZE_3(gradSecond)] = sum / (float)sumelems;
	} }
"""


def cupy_kernel(strFunction, objVariables):
    strKernel = globals()[strFunction]

    while True:
        objMatch = re.search("(SIZE_)([0-4])(\()([^\)]*)(\))", strKernel)

        if objMatch is None:
            break
        # end

        intArg = int(objMatch.group(2))

        strTensor = objMatch.group(4)
        intSizes = objVariables[strTensor].size()

        strKernel = strKernel.replace(objMatch.group(), str(intSizes[intArg]))
    # end

    while True:
        objMatch = re.search("(VALUE_)([0-4])(\()([^\)]+)(\))", strKernel)

        if objMatch is None:
            break
        # end

        intArgs = int(objMatch.group(2))
        strArgs = objMatch.group(4).split(",")

        strTensor = strArgs[0]
        intStrides = objVariables[strTensor].stride()
        strIndex = [
            "(("
            + strArgs[intArg + 1].replace("{", "(").replace("}", ")").strip()
            + ")*"
            + str(intStrides[intArg])
            + ")"
            for intArg in range(intArgs)
        ]

        strKernel = strKernel.replace(
            objMatch.group(0), strTensor + "[" + str.join("+", strIndex) + "]"
        )
    # end

    return strKernel


# end


@cupy.memoize(for_each_device=True)
def cupy_launch(strFunction, strKernel):
    return cupy.cuda.compile_with_cache(strKernel).get_function(strFunction)


# end


class _FunctionCorrelation(torch.autograd.Function):
    @staticmethod
    def forward(self, first, second):
        rbot0 = first.new_zeros(
            [first.shape[0], first.shape[2] + 8, first.shape[3] + 8, first.shape[1]]
        )
        rbot1 = first.new_zeros(
            [first.shape[0], first.shape[2] + 8, first.shape[3] + 8, first.shape[1]]
        )

        self.save_for_backward(first, second, rbot0, rbot1)

        assert first.is_contiguous()
        assert second.is_contiguous()

        output = first.new_zeros([first.shape[0], 81, first.shape[2], first.shape[3]])

        if first.is_cuda:
            n = first.shape[2] * first.shape[3]
            cupy_launch(
                "kernel_Correlation_rearrange",
                cupy_kernel(
                    "kernel_Correlation_rearrange", {"input": first, "output": rbot0}
                ),
            )(
                grid=tuple([int((n + 16 - 1) / 16), first.shape[1], first.shape[0]]),
                block=tuple([16, 1, 1]),
                args=[n, first.data_ptr(), rbot0.data_ptr()],
            )

            n = second.shape[2] * second.shape[3]
            cupy_launch(
                "kernel_Correlation_rearrange",
                cupy_kernel(
                    "kernel_Correlation_rearrange", {"input": second, "output": rbot1}
                ),
            )(
                grid=tuple([int((n + 16 - 1) / 16), second.shape[1], second.shape[0]]),
                block=tuple([16, 1, 1]),
                args=[n, second.data_ptr(), rbot1.data_ptr()],
            )

            n = output.shape[1] * output.shape[2] * output.shape[3]
            cupy_launch(
                "kernel_Correlation_updateOutput",
                cupy_kernel(
                    "kernel_Correlation_updateOutput",
                    {"rbot0": rbot0, "rbot1": rbot1, "top": output},
                ),
            )(
                grid=tuple([output.shape[3], output.shape[2], output.shape[0]]),
                block=tuple([32, 1, 1]),
                shared_mem=first.shape[1] * 4,
                args=[n, rbot0.data_ptr(), rbot1.data_ptr(), output.data_ptr()],
            )

        elif not first.is_cuda:
            raise NotImplementedError()

        # end

        return output

    # end

    @staticmethod
    def backward(self, gradOutput):
        first, second, rbot0, rbot1 = self.saved_tensors

        assert gradOutput.is_contiguous()

        gradFirst = (
            first.new_zeros(
                [first.shape[0], first.shape[1], first.shape[2], first.shape[3]]
            )
            if self.needs_input_grad[0]
            else None
        )
        gradSecond = (
            first.new_zeros(
                [first.shape[0], first.shape[1], first.shape[2], first.shape[3]]
            )
            if self.needs_input_grad[1]
            else None
        )

        if first.is_cuda:
            if gradFirst is not None:
                for intSample in range(first.shape[0]):
                    n = first.shape[1] * first.shape[2] * first.shape[3]
                    cupy_launch(
                        "kernel_Correlation_updateGradFirst",
                        cupy_kernel(
                            "kernel_Correlation_updateGradFirst",
                            {
                                "rbot0": rbot0,
                                "rbot1": rbot1,
                                "gradOutput": gradOutput,
                                "gradFirst": gradFirst,
                                "gradSecond": None,
                            },
                        ),
                    )(
                        grid=tuple([int((n + 512 - 1) / 512), 1, 1]),
                        block=tuple([512, 1, 1]),
                        args=[
                            n,
                            intSample,
                            rbot0.data_ptr(),
                            rbot1.data_ptr(),
                            gradOutput.data_ptr(),
                            gradFirst.data_ptr(),
                            None,
                        ],
                    )
                # end
            # end

            if gradSecond is not None:
                for intSample in range(first.shape[0]):
                    n = first.shape[1] * first.shape[2] * first.shape[3]
                    cupy_launch(
                        "kernel_Correlation_updateGradSecond",
                        cupy_kernel(
                            "kernel_Correlation_updateGradSecond",
                            {
                                "rbot0": rbot0,
                                "rbot1": rbot1,
                                "gradOutput": gradOutput,
                                "gradFirst": None,
                                "gradSecond": gradSecond,
                            },
                        ),
                    )(
                        grid=tuple([int((n + 512 - 1) / 512), 1, 1]),
                        block=tuple([512, 1, 1]),
                        args=[
                            n,
                            intSample,
                            rbot0.data_ptr(),
                            rbot1.data_ptr(),
                            gradOutput.data_ptr(),
                            None,
                            gradSecond.data_ptr(),
                        ],
                    )
                # end
            # end

        elif not first.is_cuda:
            raise NotImplementedError()

        # end

        return gradFirst, gradSecond

    # end


# end


def FunctionCorrelation(tenFirst, tenSecond):
    return _FunctionCorrelation.apply(tenFirst, tenSecond)


# end


class ModuleCorrelation(torch.nn.Module):
    def __init__(self):
        super(ModuleCorrelation, self).__init__()

    # end

    def forward(self, tenFirst, tenSecond):
        return _FunctionCorrelation.apply(tenFirst, tenSecond)

    # end


# end



from neurovc.contrib.flowmag.flow_models.ARFlow.losses.flow_loss import unFlowLoss


def get_loss(cfg):
    if cfg.type == "unflow":
        loss = unFlowLoss(cfg)
    else:
        raise NotImplementedError(cfg.type)
    return loss



import numpy as np
import torch
from torchvision import transforms as tf
from PIL import ImageFilter


def get_ap_transforms(cfg):
    transforms = [ToPILImage()]
    if cfg.cj:
        transforms.append(
            ColorJitter(
                brightness=cfg.cj_bri,
                contrast=cfg.cj_con,
                saturation=cfg.cj_sat,
                hue=cfg.cj_hue,
            )
        )
    if cfg.gblur:
        transforms.append(RandomGaussianBlur(0.5, 3))
    transforms.append(ToTensor())
    if cfg.gamma:
        transforms.append(RandomGamma(min_gamma=0.7, max_gamma=1.5, clip_image=True))
    return tf.Compose(transforms)


# from https://github.com/visinf/irr/blob/master/datasets/transforms.py
class ToPILImage(tf.ToPILImage):
    def __call__(self, imgs):
        return [super(ToPILImage, self).__call__(im) for im in imgs]


class ColorJitter(tf.ColorJitter):
    def __call__(self, imgs):
        transform = self.get_params(
            self.brightness, self.contrast, self.saturation, self.hue
        )
        return [transform(im) for im in imgs]


class ToTensor(tf.ToTensor):
    def __call__(self, imgs):
        return [super(ToTensor, self).__call__(im) for im in imgs]


class RandomGamma:
    def __init__(self, min_gamma=0.7, max_gamma=1.5, clip_image=False):
        self._min_gamma = min_gamma
        self._max_gamma = max_gamma
        self._clip_image = clip_image

    @staticmethod
    def get_params(min_gamma, max_gamma):
        return np.random.uniform(min_gamma, max_gamma)

    @staticmethod
    def adjust_gamma(image, gamma, clip_image):
        adjusted = torch.pow(image, gamma)
        if clip_image:
            adjusted.clamp_(0.0, 1.0)
        return adjusted

    def __call__(self, imgs):
        gamma = self.get_params(self._min_gamma, self._max_gamma)
        return [self.adjust_gamma(im, gamma, self._clip_image) for im in imgs]


class RandomGaussianBlur:
    def __init__(self, p, max_k_sz):
        self.p = p
        self.max_k_sz = max_k_sz

    def __call__(self, imgs):
        if np.random.random() < self.p:
            radius = np.random.uniform(0, self.max_k_sz)
            imgs = [im.filter(ImageFilter.GaussianBlur(radius)) for im in imgs]
        return imgs



"""
Given a test set, generates and saves predictions to a folder
Actual evaluation is done by another script
"""

import argparse
from pathlib import Path

import numpy as np
import torch
from omegaconf import OmegaConf
from torch.utils.data import DataLoader
from tqdm import tqdm

from neurovc.contrib.flowmag.dataset import get_dataloader
from neurovc.contrib.flowmag.metrics import Metrics
from neurovc.contrib.flowmag.myutils import AverageMeter, get_our_model


def test(args, model, save_dir):
    device = 0
    num_recursion = 2

    # Make place to save
    save_dir = Path(save_dir)
    save_dir.mkdir(exist_ok=True, parents=True)
    save_file = save_dir / f"flowmag_{args.flow_model}_{args.flow_model_type}.txt"
    save_file.touch()
    print(f"will be saved to {str(save_file)}")

    # Make frames dataset
    config = OmegaConf.load(args.config)
    config.config = args.config
    dataset = get_dataloader(config, "test")
    testloader = DataLoader(dataset, batch_size=1, shuffle=False)

    # Make loss function
    dotlist = [
        f"train.flow_model={args.flow_model}",
        f"train.flow_model_type={args.flow_model_type}",
        "train.raft_iters=20",
    ]
    loss_conf = OmegaConf.from_dotlist(dotlist)
    metric = Metrics(loss_conf).to(device)
    alphas = torch.tensor(args.alpha).to(device).unsqueeze(0)

    model.eval()

    print("Evaluating...")
    Motion_error = AverageMeter("motion_error")
    Mag_error = AverageMeter("mag_error")
    with torch.no_grad():
        for idx, (frames, info_data) in enumerate(tqdm(testloader)):
            config = OmegaConf.load(args.config)

            frames = frames.to(device)

            # If input alpha exceeds the range for training, perform recursion for inference
            if (
                args.alpha > config.train.alpha_high
                and np.sqrt(args.alpha) <= config.train.alpha_high
            ):
                our_alpha = np.sqrt(args.alpha)
                frames_ = frames.clone()
                for _ in range(num_recursion):
                    pred = model(frames_, alpha=our_alpha)
                    frames_ = torch.stack([frames[:, :, 0], pred[:, :, 0]], dim=2).to(
                        device
                    )

            # Run inference, if alpha is within the range of alpha for training
            elif args.alpha <= config.train.alpha_high:
                pred = model(frames, alpha=args.alpha)

            # Alpha out of range with given num of recursion
            else:
                raise Exception("alpha out of range")

            # Get metrics
            motion_error, mag_error = metric(pred, frames, alphas)

            Motion_error.update(motion_error.item())
            Mag_error.update(mag_error.item())

    print(f"Alpha: {args.alpha}")
    print(
        f"Avg motion error: {Motion_error.avg:.04f}, std: {Motion_error.std:.04f}, standard error: {Motion_error.se:.04f} @ alpha={args.alpha}"
    )
    print(
        f"Avg mag error: {Mag_error.avg:.04f}, std: {Mag_error.std:.04f}, standard error: {Mag_error.se:.04f} @ alpha={args.alpha}"
    )

    # Write in a txt file
    with open(save_file, "a+") as f:
        f.write(f"Alpha: {args.alpha}\n")
        f.write(
            f"\tAvg motion error: {Motion_error.avg:.04f}, std: {Motion_error.std:.04f}, standard error: {Motion_error.se:.04f}\n"
        )
        f.write(
            f"\tAvg mag error: {Mag_error.avg:.04f}, std: {Mag_error.std:.04f}, standard error: {Mag_error.se:.04f}\n"
        )


if __name__ == "__main__":
    # Load args and config
    parser = argparse.ArgumentParser(description="Process some integers.")
    parser.add_argument("--config", type=str, help="path to config file")
    parser.add_argument("--resume", type=str, help="path to checkpoint to resume from")
    parser.add_argument(
        "--alpha", type=float, required=True, help="amount to magnify motion"
    )
    parser.add_argument("--flow_model", type=str, default="raft")
    parser.add_argument("--flow_model_type", type=str, default="things")
    args = parser.parse_args()

    model, epoch = get_our_model(args)
    save_name = f"epoch{epoch:04}"
    config_name = args.config.split("/")[-1].split(".yaml")[-2]
    save_dir = f"./eval_results/{config_name}.ep{epoch}"

    test(args, model, save_dir)



import numpy as np
import sys
import datetime
import glob
import webbrowser
import os
import random
from neurovc.contrib.flowmag.flow_models.pwcnet.aolib import img as ig, util as ut

# PUBLIC_URL = 'http://vision13.csail.mit.edu/aho/tab'
# WWW_PATH = '/data/vision/billf/aho-vis/tables'
# PUBLIC_URL = 'http://169.229.222.157/tab'
# PUBLIC_URL = 'http://people.eecs.berkeley.edu/~owens/tab'
PUBLIC_URL = "http://owens-gpu1.eecs.umich.edu/tab"
# WWW_PATH = '/home/eecs/owens/public_html/tab'
# WWW_PATH = '/home/ao/www/tab'


def get_url():
    name = ut.hostname()
    if name == "zzz":
        return "http://128.32.43.210/tab"
    elif name in ["knish", "knodel", "knack", "kale", "kitkat"]:
        return "http://people.eecs.berkeley.edu/~owens/tab"
    elif name in ["gpu-1"]:
        return "http://people.eecs.berkeley.edu/~owens/tab"
    elif name in ["owens-gpu1"]:
        return "http://owens-gpu1.eecs.umich.edu/tab"
    elif name in ["owens-gpu2"]:
        return "http://owens-gpu2.eecs.umich.edu/tab"
    else:
        raise RuntimeError(name)


def get_www_path():
    name = ut.hostname()
    # print(name)
    if name == "zzz":
        return "/home/ao/www/tab"
    elif name in ["knish", "knodel", "knack", "kale", "kitkat"]:
        return "/home/eecs/owens/public_html/tab"
    elif name in ["gpu-1"]:
        return "../vis"
    elif name in ["owens-gpu1"]:
        # return '/home/ahowens/tab'
        return "/n/owens-data1/mnt/Big1/ahowens/tab"
    elif name in ["owens-gpu2"]:
        # return '/home/ahowens/tab'
        return "/n/owens-data1/mnt/Big1/ahowens/tab"
    else:
        raise RuntimeError(name)


def show_table(
    table,
    title=None,
    rows_per_page=100,
    show=False,
    base_dir=None,
    output_path=None,
    archive=False,
    use_www=True,
):
    """Display a table of images and text in a web browser. A table is
    represented as a list of lists, e.g. show_table([[im1, im2], ['hello', im3]]).
    Handles pagination (so you can input a table with thousands of images).

    Possible table values:
    - Arrays are assumed to be images.
    - Strings
    - A tuple ('img', filename) is assumed to be an image and its filename
    - A tuple ('img_mv', filename) is an image file that is moved into the output path
    """
    # empty html page for empty table
    # if len(table) == 0:
    #   table = ['']

    use_www = use_www and os.path.exists(get_www_path())

    if (base_dir is not None) and (not os.path.exists(base_dir)):
        print("base directory", base_dir, "does not exist", file=sys.stderr)
        base_dir = None

    # figure out the output directory from the arguments
    if output_path is None:
        # make a temporary subdirectory in base_dir if no output_path is specified
        if base_dir is None:
            # hard-coded path: since /tmp doesn't have much space on the vision machines
            if os.path.exists("/data/vision/billf/aho-vis/tmp"):
                base_dir = "/data/vision/billf/aho-vis/tmp"
            else:
                base_dir = None

        output_dir = ut.make_temp_dir(dir=base_dir)
    else:
        if base_dir is not None:
            print("Ignoring base_dir since output_path is set")
        # clear output_path and use it
        if os.path.exists(output_path):
            # todo: make this quiet when if fails
            # os.system('rm %s/*.html %s/*.png' % (output_path, output_path))
            pass
        else:
            os.system("mkdir %s" % output_path)
        output_dir = output_path

    # hard-coded path to web server
    if use_www:
        os.system("rmdir %s" % output_dir)
        symlink_dir = output_dir
        # output_dir = ut.make_temp_dir(dir = '/csail/vision-billf5/aho/www/tables')
        # public_url = 'http://quickstep.csail.mit.edu/aho/tables'
        # output_dir = ut.make_temp_dir(dir = '/csail/vision-billf5/aho/www/tab')
        output_dir = ut.make_temp_dir(dir=get_www_path())

    if title is None:
        title = datetime.datetime.today().strftime("%I:%M %p %a")

    html_rows = html_from_rows(table, output_dir)
    if use_www:
        # not sure what the minimal permissions to set are
        os.system("chmod -R a+rwx %s" % output_dir)
        os.system("ln -s %s %s" % (output_dir, symlink_dir))
        os.system("chmod -R a+rwx %s" % symlink_dir)

    fnames = paginate_table(html_rows, rows_per_page, title, output_dir)
    # adding the file:// prefix makes it possible to open the file in
    # a browser using find-file-at-point in emacs
    if use_www:
        url = os.path.join(get_url(), os.path.split(output_dir)[1])
        # symlink_fname = 'file://%s' % os.path.abspath(os.path.join(symlink_dir, 'index.html'))
        print("showing", url, "->", output_dir)
        # if show:
        #   show_page(symlink_fname)
        os.system("chmod -R a+rwx %s" % output_dir)
        return url
    else:
        print("showing", ("file://%s" % os.path.abspath(fnames[0])))
        if show:
            show_page(fnames[0])
        return fnames[0]

    if 0:
        # Save table and return HTML; don't paginate
        fname = paginate_table(html_rows, None, title, output_dir)[0]
        print("showing", ("file://%s" % os.path.abspath(fname)))

    if archive:
        dirname = os.path.split(output_dir.rstrip("/"))[1]
        zip_name = dirname + ".zip"
        # make a zip file with a single directory containing files, not a long path by changing cwd temporarily
        os.system("cd %s/..; zip -qr %s %s" % (output_dir, zip_name, dirname))
        print("saved", os.path.abspath(output_dir.rstrip("/") + ".zip"))


def table_js():
    return """
<script>
var cycle_idx = {};
var cycle_groups = {};
var cycle_ims = {};
var vid_focus = null;

// http://stackoverflow.com/questions/164397/javascript-how-do-i-print-a-message-to-the-error-console
function log(msg) {
  setTimeout(function() {
    throw new Error(msg);
  }, 0);
}

function mod(x, modulus) {
   if (x < 0) {
      x += Math.ceil(-x/modulus)*modulus;
   }
   return x % modulus;
}

function cycle(group, dim, delta, fast, ask) {
//  alert("group = " + group);
//  alert("in cycle_groups " + (group in cycle_groups));
//log('cycle called');
  var exact = -1;
  if (ask) {
    exact = parseInt(prompt("Frame number", "0"))
  } else if (fast) {
    delta *= 10;
  }
  for (var i = 0; i < cycle_groups[group].length; i++) {
    var id = cycle_groups[group][i];
    var ims = cycle_ims[id];
    cycle_idx[id][dim] = mod((exact < 0 ? cycle_idx[id][dim] + delta : exact), (dim == 0 ? ims.length : ims[cycle_idx[id][0]].length));
    //log(cycle_idx[id]);
    // clamp the other index
    cycle_idx[id][1-dim] = Math.max(0, Math.min(cycle_idx[id][1-dim], (dim == 1 ? ims.length : ims[cycle_idx[id][0]].length)));
    document.getElementById(id).src = ims[cycle_idx[id][0]][cycle_idx[id][1]];
  }
}


// the row is controlled by the number keys; column by clicking
function switch_row(e, group) {
//log('sr');
  for (var i = 0; i < cycle_groups[group].length; i++) {
     var id = cycle_groups[group][i];
     var ims = cycle_ims[id];
     // 1 - 9 ==> 0 - 8;
     var s = 48; // ascii '1'
     var n = e.charCode - s - 1;
     if (0 <= n && n < ims.length) {
        if (!(id in cycle_idx)) {
           cycle_idx[id] = [0, 0];
        }
        cycle_idx[id][0] = n;
        cycle_idx[id][1] = mod(cycle_idx[id][1], ims[n].length);
        document.getElementById(id).src = ims[cycle_idx[id][0]][cycle_idx[id][1]];
     }
   }
}

var curr_cycle_group = null;
//document.onkeypress = function(e) {
//  if (curr_cycle_group != null) {
//    switch_row(e, curr_cycle_group);
//  }
//}

document.onkeydown = function (e) {
 if (vid_focus != null && vid_focus.paused) {
    // rough estimate
    var frame_duration = 1./30;
    var vf = vid_focus;
     // 37 39 are left/right
    if (String.fromCharCode(e.keyCode) == 'O') {
      vf.currentTime = Math.max(vf.currentTime - frame_duration, 0);
    } else if (String.fromCharCode(e.keyCode) == 'P') {
      vf.currentTime = Math.min(vf.currentTime + frame_duration, vf.duration);
    }
    return true;
  } else if (curr_cycle_group != null) {
    switch_row(e, curr_cycle_group);
  }
};


function register_cycle(group, id, ims, start) {
  if (!(id in cycle_ims)) {
    if (!(group in cycle_groups)) {
      cycle_groups[group] = [];
    }
    cycle_groups[group].push(id);
    cycle_ims[id] = ims;
    cycle_idx[id] = start;
  }
}
function getParameterByName(name) {
  name = name.replace(/[\[]/, "\\\[").replace(/[\]]/, "\\\]");
  var regex = new RegExp("[\\?&]" + name + "=([^&#]*)");
  var results = regex.exec(location.search);
  return results == null ? "" : decodeURIComponent(results[1].replace(/\+/g, " "));
}

</script>
"""


def transpose_table(lsts, pad=""):
    return list(map(list, ut.zip_pad(pad, *lsts)))


def show_page(fname):
    if sys.platform == "darwin":
        # Chrome on MacOS doesn't seem to get the focus with webbrowser.open
        os.system("open %s" % fname)
    else:
        webbrowser.open(fname)


def js_list(xs):
    return "[" + ", ".join(xs) + "]"


def cycle_html(im_fnames_table, group_name, start):
    im_list_js = js_list([js_list(list(map(ut.quote, row))) for row in im_fnames_table])
    element_id = str(random.random())
    group_name = str(random.random()) if group_name is None else group_name
    default = (
        im_fnames_table[start[0]][start[1]]
        if len(im_fnames_table) and len(im_fnames_table[0])
        else ""
    )
    start_repr = repr(list(start))
    return (
        f'<img src = "{default}" id = "{element_id}" onmouseover = \'curr_cycle_group = "{group_name}";\''
        f" onclick = 'cycle(\"{group_name}\", 1, 1, event.shiftKey, event.ctrlKey)', oncontextmenu = 'cycle(\"{group_name}\", 1, -1, event.shiftKey, event.ctrlKey); return false;'"
        f' onload = \'register_cycle("{group_name}", "{element_id}", {im_list_js}, {start_repr});\'>'
    )


def paginate_table(table_rows, rows_per_page, title, output_dir):
    if rows_per_page <= 0:
        rows_per_page = int(1e100)
    split_table = list(ut.split_n(table_rows, rows_per_page))
    # html_pages = [ut.make_temp('.html', dir = output_dir) for x in split_table]
    page_names = ["index.html"] + [
        "page_%d.html" % i for i in range(2, 1 + len(split_table))
    ]
    page_paths = [os.path.join(output_dir, fname) for fname in page_names]
    for i in range(len(split_table)):
        table_html = (
            "<table border = 1><tr>" + "\n<tr>".join(split_table[i]) + "</table>"
        )
        footer = None
        if len(split_table) == 1:
            footer = ""
        else:
            footer = ""
            footer += (
                "Back " if i == 0 else "<a href = '%s'>Back</a> " % page_names[i - 1]
            )
            footer += (
                "Next "
                if i == -1 + len(page_names)
                else "<a href = '%s'>Next</a> " % page_names[i + 1]
            )
            for j in range(len(split_table)):
                s = "<b>%d</b>" % (1 + j) if (i == j) else str(1 + j)
                footer += "<a href = '%s'>%s</a> " % (page_names[j], s)
            footer += "<br><br><br><br>"
        ut.make_file(
            page_paths[i],
            "<html><head>%s<title>%s</title></head><body>%s<br>%s</html>"
            % (table_js(), title, table_html, footer),
        )
    return page_paths


def path_from_im(x, output_dir, img_encoding=".png"):
    if isinstance(x, tuple) and (x[0] in ("img", "img_mv")):
        x = list(x)
        mv_im = x[0] == "img_mv"
        if isinstance(x[1], str):
            x[1] = os.path.abspath(x[1])
        im = x[1]
    elif isinstance(x, tuple) and (x[0] in ("img-png",)):
        assert isinstance(x[1], str)
        path = ut.make_temp(".png", dir=output_dir)
        f = open(path, "wb")
        f.write(x[1])
        f.close()
        mv_im = False
        im = os.path.split(path)[1]
    else:
        mv_im = False
        im = x

    if isinstance(im, str):
        if mv_im:
            new_path = ut.make_temp(os.path.splitext(im)[1], dir=output_dir)
            # new_path = ut.make_temp(img_encoding, dir = output_dir)
            # os.rename(im, new_path)
            ut.sys_check_silent("mv", im, new_path)
            return os.path.split(new_path)[1]
        else:
            return im
    elif isinstance(im, np.ndarray):
        path = ut.make_temp(img_encoding, dir=output_dir)
        ig.save(path, im)
        # use relative path in webpage so the directory can be moved
        return os.path.split(path)[1]
    else:
        ut.fail("Don't know how to handle image type: %s" % str(type(im)))


def html_from_cell(x, output_dir):
    if isinstance(x, str):
        # text
        return x
    elif isinstance(x, np.ndarray):
        return html_from_cell(("img", x), output_dir)
    elif hasattr(x, "__video__") and x.__video__:
        return x.make_html(output_dir)
    elif hasattr(x, "__cycle__") and x.__cycle__:
        return x.make_html(output_dir)
    elif isinstance(x, tuple):
        if x[0] in ("img", "img_mv", "img-png"):
            path = path_from_im(x, output_dir)
            if len(x) >= 3 and isinstance(x[2], dict):
                opts = x[2]
            else:
                opts = dict(ut.split_n(x[2:], 2))
            # for tool-tips
            maybe_title = ('title = "%s"' % opts["title"]) if "title" in opts else ""
            html = "<img src = '%s' %s>" % (path, maybe_title)
            if "link" in opts:
                html = ut.link(opts["link"], html)
            return html
        elif x[0] == "animation":
            # make an animated gif
            duration = 0.5 if len(x) < 3 else float(x[2])
            seq_fname = make_temp_animated_gif(x[1], duration=duration, dir=output_dir)
            return html_from_cell(("img_mv", seq_fname), output_dir)
        elif x[0] == "cycle":
            # clicking: advance column; number keys: change row; shift-click: move by 10 frames; control-click: query user for column;
            im_table = x[1]
            group = None if len(x) < 3 else x[2]

            # format: [0, 0] (default) | col | [row, col]
            if len(x) < 4:
                start = [0, 0]
            elif np.ndim(x[3]) == 0:
                start = [0, x[3]]
            else:
                start = x[3]

            if len(im_table) == 0 or not isinstance(im_table[0], list):
                im_table = [im_table]

            return cycle_html(
                [[path_from_im(im, output_dir) for im in row] for row in im_table],
                group,
                start,
            )

        elif x[0] == "table":
            return "<table border = 1>%s</table>" % (
                "<tr>".join(html_from_rows(x[1], output_dir))
            )

            # ut.fail("Invalid tuple in table.")
    elif hasattr(x, "__call__"):
        # call function; make a cell from the result
        return html_from_cell(x(), output_dir)
    # # allow table of tables
    # elif type(x) == type([]):
    #   return '<table border = 1>%s</table>' % ('<tr>'.join(html_from_rows(x, output_dir)))
    else:
        return str(x)
        # ut.fail('Invalid element in table: got element of type: %s' % type(x))


# def make_temp_animated_gif(seq, duration = 0.5, dir = None):
#   from external import images2gif
#   seq_fname = ut.make_temp('.gif', dir = dir)
#   images2gif.writeGif(seq_fname, seq, duration=duration, dither=0)
#   #images2gif.writeGif(seq_fname, map(ig.to_pil, seq), duration=duration, nq = 15, subRectangles = False, dither=0)

#   return seq_fname


def save_helper(xxx_todo_changeme):
    (fname, x) = xxx_todo_changeme
    ig.save(fname, x)


# note: I think duration = #ticks before image changes, and 0.5 probably gets mapped to 0? I'm not sure.  Setting it to 1 makes it very fast, 100 much slower.
def make_temp_animated_gif_cmdline(
    seq, duration=0.5, dir=None, tmp_ext=".ppm", seq_fname=None
):
    if seq_fname is None:
        seq_fname = ut.make_temp(".gif", dir=dir)
    # avoid ffmpeg prompt
    os.system("rm %s" % seq_fname)
    base = ut.make_temp("")
    # fnames = ['%s_%d.jpg' % (base, i) for i in xrange(len(seq))]
    fnames = ["%s_%04d%s" % (base, i, tmp_ext) for i in range(len(seq))]

    # for fname, x in zip(fnames, seq):
    #   ig.save(fname, x)

    # def f((fname, x)):
    #   ig.save(fname, x)

    ut.parmap(save_helper, list(zip(fnames, seq)))

    ut.prn_sys(
        "convert -layers OptimizePlus -delay %f -loop 0 %s_*%s %s"
        % (duration, base, tmp_ext, seq_fname)
    )

    # fps = 1./duration
    # ut.prn_sys('ffmpeg -r %f -i %s_%%d.jpg %s.avi' % (fps, base, base))
    # ut.prn_sys('ffmpeg -i %s.avi -pix_fmt rgb24 %s' % (base, seq_fname))
    return seq_fname
    # os.system('rm %s*.png %s*.avi' % (base, base))


make_temp_animated_gif = make_temp_animated_gif_cmdline


def html_from_rows(table, output_dir):
    # allow single-entry table without lists
    if not isinstance(table, list):
        table = [table]
    html_rows = []
    for row in table:
        # a total hack; allows for multiple tables
        if isinstance(row, tuple) and row[0] == "break":
            html_rows.append("</table><br><table border = 1>")
        else:
            # allow single column without nesting lists
            if not isinstance(row, list):
                row = [row]
            html_rows.append(
                "<td>" + "<td>".join(html_from_cell(x, output_dir) for x in row)
            )

    return html_rows


show = show_table

# def show_links(urls):
#   table = []
#   for url in urls:
#     if ut.istup(url):
#       table.append


def test():
    # show([[['hello', 'world'],
    #        'asdf', 'fdsa',
    #        ['a', 'b', 'c', 'd', 'e', 'f']],
    #       [['foo', 'bar'], ['x']]])
    show(
        [
            (
                "table",
                [["hello", "world"], "asdf", "fdsa", ["a", "b", "c", "d", "e", "f"]],
            ),
            ("table", [["foo", "bar"], ["x"]]),
        ]
    )

    # animated gif test
    show(("animation", [ig.make(50, 50), ig.make(50, 50, (255, 0, 0))], 0.1))


def pad_im_even(x):
    if x.shape[0] % 2 == 1 or x.shape[1] % 2 == 1:
        x = ig.pad_corner(x, x.shape[1] % 2, x.shape[0] % 2)
    return x


def make_video_helper(xxx_todo_changeme1):
    (i, x, in_dir, tmp_ext) = xxx_todo_changeme1
    out_fname = ut.pjoin(in_dir, "ao-video-frame%05d%s" % (i, tmp_ext))
    if isinstance(x, str):
        f = open(out_fname, "wb")
        f.write(x)
        f.close()
    else:
        # pad to be even
        x = pad_im_even(x)
        ig.save(out_fname, x)


# def make_video(out_fname, ims, fps, tmp_ext = '.ppm', sound = None):
#   # Compressed images
#   if type(ims[0]) == type(''):
#     #tmp_ext = '.jpg'
#     tmp_ext = '.png'

#   assert tmp_ext.startswith('.')
#   in_dir = ut.make_temp_dir_big()
#   ut.parmap(make_video_helper,
#             [(i, x, in_dir, tmp_ext) for i, x in enumerate(ims)])

#   with ut.temp_file('.wav') as tmp_wav:
#     if sound is None:
#       sound_flags = ''
#     else:
#       if type(sound) != type(''):
#         sound.save(tmp_wav)
#         sound = tmp_wav
#       sound_flags = '-i "%s" -shortest' % sound

#     cmd = ('ffmpeg -loglevel warning -f image2 -r %f -i %s/ao-video-frame%%05d%s %s '
#            '-pix_fmt yuv420p -vcodec h264 -acodec aac -strict -2 -y %s') % (fps, in_dir, tmp_ext, sound_flags, out_fname)

#     print cmd
#     if 0 == os.system(cmd):
#       for x in glob.glob(ut.pjoin(in_dir, 'ao-video-frame*%s' % tmp_ext)):
#         os.remove(x)
#       os.rmdir(in_dir)
#     else:
#       raise RuntimeError()


def make_video(out_fname, ims, fps, tmp_ext=".ppm", sound=None):
    # Compressed images
    if isinstance(ims[0], str):
        # tmp_ext = '.jpg'
        tmp_ext = ".png"

    assert tmp_ext.startswith(".")
    in_dir = ut.make_temp_dir_big()
    ut.parmap(make_video_helper, [(i, x, in_dir, tmp_ext) for i, x in enumerate(ims)])

    with ut.temp_file(".wav") as tmp_wav:
        if sound is None:
            sound_flags = ""
        else:
            if not isinstance(sound, str):
                sound.save(tmp_wav)
                sound = tmp_wav
            sound_flags = '-i "%s" -shortest' % sound

        # removed aac flag from above, for compatibility with some wav files
        # cmd = ('ffmpeg -loglevel warning -f image2 -r %f -i %s/ao-video-frame%%05d%s %s '
        #        '-pix_fmt yuv420p -vcodec h264 -strict -2 -y %s') % (fps, in_dir, tmp_ext, sound_flags, out_fname)
        cmd = (
            "ffmpeg -loglevel error -f image2 -r %f -i %s/ao-video-frame%%05d%s %s "
            "-pix_fmt yuv420p -ar 22050 -vcodec h264 -strict -2 -y %s"
        ) % (fps, in_dir, tmp_ext, sound_flags, out_fname)

        print(cmd)
        if 0 == os.system(cmd):
            for x in glob.glob(ut.pjoin(in_dir, "ao-video-frame*%s" % tmp_ext)):
                os.remove(x)
            os.rmdir(in_dir)
        else:
            raise RuntimeError()


def convert_video(in_fname, out_fname):
    # cmd = ('ffmpeg -loglevel warning -i "%s" '
    #        '-pix_fmt yuv420p -vcodec h264 -acodec aac -strict -2 -y %s') % (in_fname, out_fname)
    cmd = (
        'ffmpeg -loglevel fatal -i "%s" '
        "-pix_fmt yuv420p -vcodec h264  -strict -2 -y %s"
    ) % (in_fname, out_fname)
    print(cmd)
    os.system(cmd)


def test_video_cycle():
    ims = [
        "/afs/csail.mit.edu/u/a/aho/bear.jpg",
        "/afs/csail.mit.edu/u/a/aho/test.jpg",
        "/afs/csail.mit.edu/u/a/aho/me_cropped.jpg",
    ]
    ims = [ig.load(x)[:120, :120] for x in ims]
    show([Video(ims)])
    print("cycle")
    show([Cycle(ims, None)])
    # show([Video(map(ig.load, ims))])


def sync_with_fps(ims, timestamps, target_fps, max_lag_sec=0.06):
    vid = []
    timestamps = np.array(timestamps, "d")
    # timestamps -= timestamps[0]
    i = 0
    while i < len(ims):
        vid.append(ims[i])

        def video_time():
            return (len(vid) - 1) * (1.0 / target_fps)

        # print video_time(), timestamps[i]
        # We're over time; skip frames until we catch up with the timestamps
        while i < len(timestamps) and video_time() > timestamps[i] + max_lag_sec:
            print("Over time. Skipping frames to catch up.")
            i += 1

        # We're far ahead of the true time.  Keep showing the same image until we reach it again.
        while video_time() < timestamps[i] - max_lag_sec:
            print("Under time. Repeating the last frame to slow down.")
            vid.append(ims[i])

        i += 1

    return vid


class Video:
    def __init__(
        self,
        ims=None,
        fps=1,
        sound=None,
        timestamps=None,
        video_fname=None,
        copy_method="copy",
        video_url=None,
    ):
        self.__video__ = True
        self.video_fname = video_fname
        self.video_url = video_url
        if video_fname is not None:
            self.copy_method = copy_method
            # return
        # print 'init video'
        if timestamps is None:
            self.ims = ims
        else:
            self.ims = sync_with_fps(ims, timestamps, fps)

        self.fps = fps
        self.sound = sound

    def make_html(self, out_dir, ext=".mp4"):
        if self.video_url is None:
            if self.video_fname is None:
                fname = ut.make_temp(ext, dir=out_dir)
                os.remove(fname)
                make_video(fname, self.ims, self.fps, sound=self.sound)
            else:
                fname = ut.make_temp(ext, dir=out_dir)
                os.remove(fname)
                if self.copy_method == "copy":
                    os.system('cp "%s" %s' % (self.video_fname, fname))
                elif self.copy_method == "symlink":
                    os.system('ln -s "%s" %s' % (self.video_fname, fname))
                else:
                    raise RuntimeError()
            rel_fname = fname.split("/")[-1]
        else:
            rel_fname = self.video_url

        vid_id = str(random.random())

        return """
    <table>
      <tr>
        <td>
          <video controls id = "%(vid_id)s" onplay = "vid_focus = null;" onpause = "vid_focus = document.getElementById('%(vid_id)s');" onloadedmetadata = "var t = getParameterByName('videoStart'); if (t != null) { this.currentTime = t; } if (getParameterByName('videoAutoplay')) { this.play(); }">
          <source src = "%(rel_fname)s" type = "video/mp4" preload = "none"> </video>
        </td>
        <td>
          Speed: <input type = "text" size = "5" value = "1" oninput = "document.getElementById('%(vid_id)s').playbackRate = parseFloat(event.target.value);">
        </td>
      </tr>
    </table>
          """ % locals()

    def save(self, fname):
        assert fname.endswith(".mp4")
        if os.path.exists(fname):
            os.remove(fname)
        make_video(fname, self.ims, self.fps, sound=self.sound)


class Cycle:
    def __init__(self, ims, group=None):
        self.ims = ims
        self.group = group
        self.start = [0, 0]
        self.__cycle__ = True

    def make_html(self, output_dir):
        if len(self.ims) == 0 or not isinstance(self.ims[0], list):
            ims = [self.ims]

        return cycle_html(
            [[path_from_im(im, output_dir) for im in row] for row in ims],
            self.group,
            self.start,
        )


# def url_from_path(path):
#   path.split('/'
#   path.replace(get_www_path(),



import types
from pathlib import Path

from neurovc.thermal_landmarks import tfw_util


def test_model_downloader_uses_gdown(monkeypatch, tmp_path):
    calls = []

    def fake_download(url, output, quiet=False):
        calls.append({"url": url, "output": Path(output), "quiet": quiet})
        Path(output).write_bytes(b"weights")
        return output

    fake_gdown = types.SimpleNamespace(download=fake_download)

    def fake_require(module: str, *, extra: str, purpose: str):
        assert module == "gdown"
        return fake_gdown

    monkeypatch.setattr(tfw_util, "require", fake_require)

    downloader = tfw_util._ModelDownloader("YOLOv5n-Face.modern", save_dir=tmp_path)
    model_path = downloader.download_model()

    expected_model_path = tmp_path / tfw_util._file_targets["YOLOv5n-Face.modern"]
    expected_license_path = tmp_path / tfw_util._file_targets["license"]

    assert model_path == expected_model_path
    assert model_path.exists()
    assert expected_license_path.exists()

    assert len(calls) == 2
    model_call = next(call for call in calls if call["output"] == expected_model_path)
    license_call = next(
        call for call in calls if call["output"] == expected_license_path
    )

    assert tfw_util._file_id_map["YOLOv5n-Face.modern"] in model_call["url"]
    assert tfw_util._file_id_map["license"] in license_call["url"]


def test_model_downloader_skips_existing_file(monkeypatch, tmp_path):
    existing = tmp_path / tfw_util._file_targets["YOLOv5n-Face.modern"]
    existing.write_bytes(b"preexisting")
    existing_license = tmp_path / tfw_util._file_targets["license"]
    existing_license.write_text("preexisting license", encoding="utf-8")

    def fail_require(*args, **kwargs):  # pragma: no cover - would signal a bug
        raise AssertionError("require() should not be called when files exist")

    monkeypatch.setattr(tfw_util, "require", fail_require)

    downloader = tfw_util._ModelDownloader("YOLOv5n-Face.modern", save_dir=tmp_path)
    model_path = downloader.download_model()

    assert model_path == existing
    assert model_path.read_bytes() == b"preexisting"
    assert existing_license.read_text(encoding="utf-8") == "preexisting license"


def test_tfw_landmarker_init_loads_model(monkeypatch, tmp_path):
    model_path = tmp_path / "dummy.pt"
    model_path.write_bytes(b"stub")

    calls = {}

    def fake_prepare_model(model_name="YOLOv5n-Face.modern"):
        calls["prepare_model"] = model_name
        return model_path

    monkeypatch.setattr(tfw_util, "_prepare_model", fake_prepare_model)

    class DummyCuda:
        @staticmethod
        def is_available():
            return False

    def fake_load_model(path, device=None):
        calls["load_model"] = {"path": Path(path), "device": device}
        return "MODEL"

    dummy_torch = types.SimpleNamespace(device=lambda value: value, cuda=DummyCuda)
    dummy_yf = types.SimpleNamespace(load_model=fake_load_model)

    def fake_require(module: str, *, extra: str, purpose: str):
        calls.setdefault("require", []).append((module, extra, purpose))
        if module == "torch":
            return dummy_torch
        if module == "yolov5_face.detect_face":
            return dummy_yf
        raise AssertionError(f"Unexpected require() call: {module}")

    monkeypatch.setattr(tfw_util, "require", fake_require)

    landmarker = tfw_util.TFWLandmarker(model_name="YOLOv5n-Face.modern")

    assert calls["prepare_model"] == "YOLOv5n-Face.modern"
    assert calls["load_model"]["path"] == model_path
    assert calls["load_model"]["device"] is None
    assert landmarker.device is None
    assert landmarker.model == "MODEL"



import os
import json
from pathlib import Path

import numpy as np
from PIL import Image

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data import Dataset
from torchvision.transforms.functional import to_tensor, resize
from torchvision.transforms import (
    RandomResizedCrop,
    Compose,
    RandomHorizontalFlip,
    RandomVerticalFlip,
    RandomRotation,
    CenterCrop,
    Resize,
    ColorJitter,
)

from einops import rearrange


class RepeatDataset(Dataset):
    def __init__(self, dataset, factor=10):
        self.dataset = dataset
        self.factor = factor

    def __getitem__(self, idx):
        idx = idx % len(self.dataset)
        return self.dataset[idx]

    def __len__(self):
        return self.factor * len(self.dataset)


class TrainingFramesDataset(Dataset):
    def __init__(self, root):
        self.root = Path(root)
        self.frame_names = sorted(os.listdir(self.root))

    def transform_frame(self, frame):
        # Crop to H and W multiple of 8 (for RAFT)
        # n.b. technically this is only necessary for training
        c, h, w = frame.shape
        frame = frame[:, : (h // 8) * 8, : (w // 8) * 8]
        return frame

    def __getitem__(self, idx):
        fname = self.frame_names[idx]

        frame = Image.open(self.root / fname)
        frame = to_tensor(frame)

        # # Transform tensors
        frame = self.transform_frame(frame)

        return frame

    def __len__(self):
        return len(self.frame_names)


class FramesDataset(Dataset):
    def __init__(self, root):
        self.root = Path(root)
        self.frame_names = sorted(os.listdir(self.root))

    def transform_frame(self, frame):
        # Crop to H and W multiple of 8 (for RAFT)
        # n.b. technically this is only necessary for training
        c, h, w = frame.shape
        frame = frame[:, : (h // 8) * 8, : (w // 8) * 8]
        return frame

    def __getitem__(self, idx):
        fname = self.frame_names[idx]

        frame = Image.open(self.root / fname)
        frame = to_tensor(frame)

        return frame

    def __len__(self):
        return len(self.frame_names)


class TestTimeAdaptDataset(Dataset):
    def __init__(self, root, mode="first", length=None):
        """
        args:
            root: (string) path to directory of frames
            mode: ['first', 'random'] how to sample frames
                first: always samples first frame + idx^th frame
                random: randomly samples two frames
        """
        self.root = Path(root)
        self.frame_names = sorted(os.listdir(self.root))
        self.mode = mode
        self.im_size = 512
        self.scale = 1.1  # stretching scale

        self.geom_transform = Compose(
            [
                RandomRotation(5),
            ]
        )
        self.color_transform = ColorJitter(
            brightness=0.5, contrast=0.5, saturation=0.5, hue=0.3
        )

        if length is not None:
            self.length = length
        else:
            self.length = len(self.frame_names)

    def transform_frames(self, frames):
        c, t, h, w = frames.shape

        # Apply geometric transforms on all frames
        frames = rearrange(frames, "c t h w -> (c t) h w")
        frames = self.geom_transform(frames)
        frames = rearrange(frames, "(c t) h w -> c t h w", c=c, t=t)

        # scale
        new_h = h * self.scale ** np.random.uniform(-1, 1)  # rescale for stretch aug
        new_w = w * self.scale ** np.random.uniform(-1, 1)
        new_h = int(new_h)
        new_w = int(new_w)
        frames = resize(frames, (new_h, new_w))

        # Pad out so edges are at least im_size.
        to_pad_h = max(self.im_size - new_h, 0)
        to_pad_w = max(self.im_size - new_w, 0)
        pad_l_h = to_pad_h // 2
        pad_r_h = to_pad_h - pad_l_h
        pad_l_w = to_pad_w // 2
        pad_r_w = to_pad_w - pad_l_w
        frames = F.pad(frames, (pad_l_w, pad_r_w, pad_l_h, pad_r_h))
        _, _, padded_h, padded_w = frames.shape

        # crop to (im_size, im_size)
        ch = self.im_size
        cw = self.im_size
        ct = np.random.randint(padded_h - ch + 1)
        cl = np.random.randint(padded_w - cw + 1)
        frames = frames[:, :, ct : ct + ch, cl : cl + cw]

        # Apply color transforms (must have c=3, treating t as batch)
        frames = rearrange(frames, "c t h w -> t c h w")
        frames = self.color_transform(frames)
        frames = rearrange(frames, "t c h w -> c t h w")

        return frames

    def __getitem__(self, idx):
        idx = idx % len(self.frame_names)

        if self.mode == "first":
            idx0 = 0
            idx1 = idx
        elif self.mode == "random":
            idx0 = np.random.randint(self.length)
            idx1 = np.random.randint(self.length)
        else:
            raise NotImplementedError

        frame0 = Image.open(self.root / self.frame_names[idx0])
        frame1 = Image.open(self.root / self.frame_names[idx1])

        frame0 = to_tensor(frame0)
        frame1 = to_tensor(frame1)

        frames = torch.stack([frame0, frame1], dim=1)

        # Transform tensors
        frames = self.transform_frames(frames)

        return frames

    def __len__(self):
        return self.length


class FlowMagDataset(Dataset):
    def __init__(self, data_root, split, aug=False, img_size=256):
        self.data_root = Path(data_root)
        self.split = split
        self.img_size = img_size

        # There is not valid set, only a test set
        if self.split == "valid":
            self.split = "test"

        # Get frame metadata
        self.frameA_dir = self.data_root / self.split / "frameA"
        self.frameB_dir = self.data_root / self.split / "frameB"
        with open(self.data_root / f"{self.split}_fn.json", "r") as f:
            self.fnames = json.load(f)

        # Make augmentations
        if aug:
            self.transform = Compose(
                [
                    RandomResizedCrop(img_size, scale=(0.7, 1.0)),
                    RandomHorizontalFlip(0.5),
                    RandomVerticalFlip(0.5),
                    RandomRotation(15),
                ]
            )
            self.color_transform = ColorJitter(
                brightness=0.5, contrast=0.5, saturation=0.5, hue=0.3
            )
        else:
            self.transform = Compose(
                [
                    Resize(img_size),
                    CenterCrop(img_size),
                ]
            )
            self.color_transform = nn.Identity()

    def transform_frames(self, frames):
        c, t, h, w = frames.shape

        # Apply geometric transforms on all frames
        frames = rearrange(frames, "c t h w -> (c t) h w")
        frames = self.transform(frames)
        frames = rearrange(frames, "(c t) h w -> c t h w", c=c, t=t)

        # Apply color transforms (must have c=3, treating t as batch)
        frames = rearrange(frames, "c t h w -> t c h w")
        frames = self.color_transform(frames)
        frames = rearrange(frames, "t c h w -> c t h w")

        return frames

    def __getitem__(self, idx):
        # Load both frameA and frameB
        image_paths = [
            self.frameA_dir / self.fnames[idx],
            self.frameB_dir / self.fnames[idx],
        ]
        images = [Image.open(path) for path in image_paths]
        images = [to_tensor(im) for im in images]
        frames = torch.stack(images, dim=1)

        frames = self.transform_frames(frames)

        info = {"fname": self.fnames[idx]}

        return frames, info

    def __len__(self):
        return len(self.fnames)


def get_dataloader(config, split):
    if split == "train":
        aug = config.data.aug
    else:
        aug = False

    dataset = FlowMagDataset(
        config.data.dataroot, split=split, aug=aug, img_size=config.data.im_size
    )

    return dataset



import cv2
import numpy as np
import torch
import torch.nn as nn
from einops import rearrange
from omegaconf import OmegaConf
from torchvision.utils import make_grid, save_image

from neurovc.contrib.flowmag.model import MotionMagModel


class AverageMeter(object):
    def __init__(self, name):
        self.name = name
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0
        self.data = []

    def update(self, val, n=1):
        # Compute the sum, avg, std and standard error for data
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count
        self.data.append(val)
        self.std = np.std(self.data)
        self.se = np.std(self.data, ddof=1) / np.sqrt(np.size(self.data))

    def __str__(self):
        return f"{self.name}: {self.val:.5f} {self.avg:.5f}"


def write_video(frames, fps, output_path):
    # Write a list of array for frames into mp4 files
    h, w, _ = frames[0].shape
    fourcc = cv2.VideoWriter_fourcc("m", "p", "4", "v")
    writer = cv2.VideoWriter(output_path, fourcc, fps, (w, h))
    for frame in frames:
        writer.write(frame)

    writer.release()


def get_our_model(args, training=False):
    # Get our model for inference or evaluation
    config = OmegaConf.load(args.config)
    config.config = args.config

    # Force configs
    config.train.ngpus = 1
    config.train.is_training = True if training else False
    config.data.batch_size = 1

    # Make model
    print("Making model")
    device = 0
    dp_devices = list(range(1))
    model = MotionMagModel(config).to(device)
    model = nn.DataParallel(model, device_ids=dp_devices)

    # Resume model
    print(f"Resuming from {args.resume}")
    chkpt = torch.load(args.resume)
    model.load_state_dict(chkpt["state_dict"], strict=False)

    return model, chkpt["epoch"]


def dist_transform(mask):
    h, w = mask.shape

    # get closest in south east
    def closest_se(mask):
        closest = torch.ones(mask.shape) * float("inf")
        for i in range(1, h):
            for j in range(1, w):
                if mask[i, j] == 1:
                    closest[i, j] = 0
                else:
                    closest[i, j] = min(closest[i, j - 1], closest[i - 1, j]) + 1

        return closest

    # Get all four possible directions
    se = closest_se(mask)
    nw = closest_se(mask.flip((0, 1))).flip((0, 1))
    sw = closest_se(mask.flip(0)).flip(0)
    ne = closest_se(mask.flip(1)).flip(1)

    res_0 = torch.minimum(se, nw)
    res_1 = torch.minimum(sw, ne)
    res = torch.minimum(res_0, res_1)

    return res


def log_images(writer, images_dict, epoch_num, split, config):
    # save images
    save_dir = config.save_dir / "images"
    save_dir.mkdir(exist_ok=True, parents=True)

    for name, images in images_dict.items():
        for idx, image in enumerate(rearrange(images, "b c t h w -> t b c h w")):
            grid = make_grid(image)
            writer.add_image(f"{name}_{idx}/{split}", grid, epoch_num)
            save_image(
                grid, save_dir / f"epoch{epoch_num:05}.{split}.{name}{idx:02}.png"
            )



"""FlowMag model download utilities."""

from pathlib import Path
from typing import Dict

from neurovc.contrib.flowmag import FLOWMAG_MODEL_FILE_IDS

_FILE_TARGETS: Dict[str, str] = {
    "raft": "raft_chkpt_00140.pth",
    "arflow": "arflow_chkpt_00140.pth",
}


class _ModelDownloader:
    def __init__(self, model_name: str, save_dir: Path | str | None = None):
        self.model_name = model_name
        self.file_id = FLOWMAG_MODEL_FILE_IDS.get(model_name)
        if self.file_id is None:
            raise ValueError(
                f"Model name '{model_name}' is not valid. "
                f"Available: {sorted(FLOWMAG_MODEL_FILE_IDS)}"
            )

        base_dir = (
            Path(save_dir)
            if save_dir is not None
            else Path("~/.neurovc/models/flowmag")
        )
        self.save_dir = base_dir.expanduser()
        target = _FILE_TARGETS.get(model_name, f"{model_name}.pth")
        self.model_path = self.save_dir / target

    def download_model(self) -> Path:
        self.save_dir.mkdir(parents=True, exist_ok=True)

        if self.model_path.exists():
            return self.model_path

        try:
            import gdown
        except ImportError as exc:  # pragma: no cover - import guard
            raise ImportError(
                "gdown is required to download FlowMag checkpoints. "
                "Install it via 'pip install gdown'."
            ) from exc

        url = f"https://drive.google.com/uc?export=download&id={self.file_id}"
        gdown.download(url, str(self.model_path), quiet=False)
        return self.model_path


def download_flowmag_model(model_name: str, save_dir: Path | str | None = None) -> Path:
    """Download a single FlowMag checkpoint by model name and return its path."""
    return _ModelDownloader(model_name=model_name, save_dir=save_dir).download_model()


def download_all_flowmag_models(save_dir: Path | str | None = None) -> Dict[str, Path]:
    """Download all known FlowMag checkpoints and return their paths keyed by name."""
    return {
        name: download_flowmag_model(name, save_dir=save_dir)
        for name in FLOWMAG_MODEL_FILE_IDS
    }


# Pre-configured convenience instance mirroring tfake usage style
flowmag_model_downloader = _ModelDownloader

__all__ = [
    "_ModelDownloader",
    "download_flowmag_model",
    "download_all_flowmag_models",
    "flowmag_model_downloader",
]



"""
Minimal setup.py for backwards compatibility and editable installs.
All configuration is in pyproject.toml.
"""

from setuptools import setup

# This enables:
# - pip install .
# - pip install -e .
# - python setup.py develop (deprecated but still used)
#
# All actual configuration comes from pyproject.toml
setup()



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
# https://github.com/facebookresearch/detr/blob/main/models/position_encoding.py

import torch
import torch.nn as nn
import math


class PositionEmbeddingSine(nn.Module):
    """
    This is a more standard version of the position embedding, very similar to the one
    used by the Attention is all you need paper, generalized to work on images.
    """

    def __init__(self, num_pos_feats=64, temperature=10000, normalize=True, scale=None):
        super().__init__()
        self.num_pos_feats = num_pos_feats
        self.temperature = temperature
        self.normalize = normalize
        if scale is not None and normalize is False:
            raise ValueError("normalize should be True if scale is passed")
        if scale is None:
            scale = 2 * math.pi
        self.scale = scale

    def forward(self, x):
        # x = tensor_list.tensors  # [B, C, H, W]
        # mask = tensor_list.mask  # [B, H, W], input with padding, valid as 0
        b, c, h, w = x.size()
        mask = torch.ones((b, h, w), device=x.device)  # [B, H, W]
        y_embed = mask.cumsum(1, dtype=torch.float32)
        x_embed = mask.cumsum(2, dtype=torch.float32)
        if self.normalize:
            eps = 1e-6
            y_embed = y_embed / (y_embed[:, -1:, :] + eps) * self.scale
            x_embed = x_embed / (x_embed[:, :, -1:] + eps) * self.scale

        dim_t = torch.arange(self.num_pos_feats, dtype=torch.float32, device=x.device)
        dim_t = self.temperature ** (2 * (dim_t // 2) / self.num_pos_feats)

        pos_x = x_embed[:, :, :, None] / dim_t
        pos_y = y_embed[:, :, :, None] / dim_t
        pos_x = torch.stack(
            (pos_x[:, :, :, 0::2].sin(), pos_x[:, :, :, 1::2].cos()), dim=4
        ).flatten(3)
        pos_y = torch.stack(
            (pos_y[:, :, :, 0::2].sin(), pos_y[:, :, :, 1::2].cos()), dim=4
        ).flatten(3)
        pos = torch.cat((pos_y, pos_x), dim=3).permute(0, 3, 1, 2)
        return pos



__author__ = "Philipp Flotho"
"""
multimodal_cam_calib
Copyright 2021 by Philipp Flotho, All rights reserved.
"""
import numpy as np  # noqa: E402
import cv2  # noqa: E402

from neurovc.util.IO_util import imagesc  # noqa: E402
from neurovc.util.IO_util import normalize_color  # noqa: E402


def _require_h5py(purpose: str):
    try:
        import h5py  # type: ignore
    except ImportError as exc:
        raise ImportError(
            f"h5py is required for {purpose}. "
            "Install via `pip install h5py` or `pip install neurovc[momag]`."
        ) from exc
    return h5py


class SingleRecording:
    # class works for hfr and thermal hdf recordings
    def __init__(self, hdf_file, n_frames=None, shape=None):
        self.hdf_file = hdf_file
        self._h5py = _require_h5py("SingleRecording")

        with self._h5py.File(self.hdf_file, "r") as f:
            frame_key = "Frames" if "Frames" in f.keys() else "FRAMES"
            self.width = f[frame_key].shape[1]  # .attrs['FrameWidth']
            self.height = f[frame_key].shape[2]  # .attrs['FrameHeight']
            if n_frames is None:
                self.n_frames = f[frame_key].shape[0]  # .attrs['FrameCount']
            else:
                self.n_frames = n_frames

        self.frames = self._h5py.File(self.hdf_file, "r")[frame_key]
        self.timestamps = self._h5py.File(self.hdf_file, "r")["Timestamps_ms"][:]
        self.shape = shape

    def load_data(self):
        pass

    def get_frames(self, idx):
        if self.shape is None:
            frames = self.frames[idx]
        else:
            frames = self.frames[idx, : self.shape[0], : self.shape[1]]
        timestamps = self.timestamps[idx]
        return frames, timestamps

    def plot_thermal_frame(self, frame_num):
        # resize the thermal image or representation
        test_frame = np.resize(self.frames[frame_num], (480, 640))
        # norm the thermal image
        norm_img = normalize_color(test_frame)
        imagesc(norm_img)

    def get_thermal_frame(self, frame_num):
        thermal_frame = np.resize(self.frames[frame_num], (480, 640))

        return thermal_frame


def __test_thermal2():
    # thermal_file = "/home/philipp/covid/Thermal/2020_05_07/ThermalData_07_05_2020_10_38_43.h5"
    thermal_file = (
        "G:/covid-19/data/pilot3/elena/thermal/ThermalData_18_06_2020_13_24_58.h5"
    )
    tmp = SingleRecording(thermal_file, n_frames=10)
    tmp.load_data()
    imagesc(tmp.frames[0])
    cv2.waitKey()


def synchronize_timestamps(left_ts, right_ts, fps=30):
    timestamp_baseline = np.min(np.array([left_ts[0], right_ts[0]]))

    left_ts = np.array(left_ts[:] - timestamp_baseline).astype(float)
    right_ts = np.array(right_ts[:] - timestamp_baseline).astype(float)

    start_time = 0
    end_time = np.min(np.array([left_ts[-1], right_ts[-1]]))

    left_time_resampled = np.arange(start_time, end_time, 1000 / fps, float)
    idx_left = np.round(
        np.interp(left_time_resampled, left_ts, np.arange(0, left_ts.shape[0]))
    ).astype(int)
    right_time_resampled = np.arange(start_time, end_time, 1000 / fps, float)
    idx_right = np.round(
        np.interp(right_time_resampled, right_ts, np.arange(0, right_ts.shape[0]))
    ).astype(int)

    return idx_left, left_time_resampled, idx_right, right_time_resampled


if __name__ == "__main__":
    __test_thermal2()



import torch
import torch.nn as nn
import inspect


def mesh_grid(B, H, W):
    # mesh grid
    x_base = torch.arange(0, W).repeat(B, H, 1)  # BHW
    y_base = torch.arange(0, H).repeat(B, W, 1).transpose(1, 2)  # BHW

    base_grid = torch.stack([x_base, y_base], 1)  # B2HW
    return base_grid


def norm_grid(v_grid):
    _, _, H, W = v_grid.size()

    # scale grid to [-1,1]
    v_grid_norm = torch.zeros_like(v_grid)
    v_grid_norm[:, 0, :, :] = 2.0 * v_grid[:, 0, :, :] / (W - 1) - 1.0
    v_grid_norm[:, 1, :, :] = 2.0 * v_grid[:, 1, :, :] / (H - 1) - 1.0
    return v_grid_norm.permute(0, 2, 3, 1)  # BHW2


def get_corresponding_map(data):
    """

    :param data: unnormalized coordinates Bx2xHxW
    :return: Bx1xHxW
    """
    B, _, H, W = data.size()

    # x = data[:, 0, :, :].view(B, -1).clamp(0, W - 1)  # BxN (N=H*W)
    # y = data[:, 1, :, :].view(B, -1).clamp(0, H - 1)

    x = data[:, 0, :, :].view(B, -1)  # BxN (N=H*W)
    y = data[:, 1, :, :].view(B, -1)

    # invalid = (x < 0) | (x > W - 1) | (y < 0) | (y > H - 1)   # BxN
    # invalid = invalid.repeat([1, 4])

    x1 = torch.floor(x)
    x_floor = x1.clamp(0, W - 1)
    y1 = torch.floor(y)
    y_floor = y1.clamp(0, H - 1)
    x0 = x1 + 1
    x_ceil = x0.clamp(0, W - 1)
    y0 = y1 + 1
    y_ceil = y0.clamp(0, H - 1)

    x_ceil_out = x0 != x_ceil
    y_ceil_out = y0 != y_ceil
    x_floor_out = x1 != x_floor
    y_floor_out = y1 != y_floor
    invalid = torch.cat(
        [
            x_ceil_out | y_ceil_out,
            x_ceil_out | y_floor_out,
            x_floor_out | y_ceil_out,
            x_floor_out | y_floor_out,
        ],
        dim=1,
    )

    # encode coordinates, since the scatter function can only index along one axis
    corresponding_map = torch.zeros(B, H * W).type_as(data)
    indices = torch.cat(
        [
            x_ceil + y_ceil * W,
            x_ceil + y_floor * W,
            x_floor + y_ceil * W,
            x_floor + y_floor * W,
        ],
        1,
    ).long()  # BxN   (N=4*H*W)
    values = torch.cat(
        [
            (1 - torch.abs(x - x_ceil)) * (1 - torch.abs(y - y_ceil)),
            (1 - torch.abs(x - x_ceil)) * (1 - torch.abs(y - y_floor)),
            (1 - torch.abs(x - x_floor)) * (1 - torch.abs(y - y_ceil)),
            (1 - torch.abs(x - x_floor)) * (1 - torch.abs(y - y_floor)),
        ],
        1,
    )
    # values = torch.ones_like(values)

    values[invalid] = 0

    corresponding_map.scatter_add_(1, indices, values)
    # decode coordinates
    corresponding_map = corresponding_map.view(B, H, W)

    return corresponding_map.unsqueeze(1)


def flow_warp(x, flow12, pad="border", mode="bilinear"):
    B, _, H, W = x.size()

    base_grid = mesh_grid(B, H, W).type_as(x)  # B2HW

    v_grid = norm_grid(base_grid + flow12)  # BHW2
    if "align_corners" in inspect.getfullargspec(torch.nn.functional.grid_sample).args:
        im1_recons = nn.functional.grid_sample(
            x, v_grid, mode=mode, padding_mode=pad, align_corners=True
        )
    else:
        im1_recons = nn.functional.grid_sample(x, v_grid, mode=mode, padding_mode=pad)
    return im1_recons


def get_occu_mask_bidirection(flow12, flow21, scale=0.01, bias=0.5):
    flow21_warped = flow_warp(flow21, flow12, pad="zeros")
    flow12_diff = flow12 + flow21_warped
    mag = (flow12 * flow12).sum(1, keepdim=True) + (flow21_warped * flow21_warped).sum(
        1, keepdim=True
    )
    occ_thresh = scale * mag + bias
    occ = (flow12_diff * flow12_diff).sum(1, keepdim=True) > occ_thresh
    return occ.float()


def get_occu_mask_backward(flow21, th=0.2):
    B, _, H, W = flow21.size()
    base_grid = mesh_grid(B, H, W).type_as(flow21)  # B2HW

    corr_map = get_corresponding_map(base_grid + flow21)  # BHW
    occu_mask = corr_map.clamp(min=0.0, max=1.0) < th
    return occu_mask.float()



import torch.nn as nn

from neurovc.contrib.flowmag.flow_models.gmflow.trident_conv import (
    MultiScaleTridentConv,
)


class ResidualBlock(nn.Module):
    def __init__(
        self,
        in_planes,
        planes,
        norm_layer=nn.InstanceNorm2d,
        stride=1,
        dilation=1,
    ):
        super(ResidualBlock, self).__init__()

        self.conv1 = nn.Conv2d(
            in_planes,
            planes,
            kernel_size=3,
            dilation=dilation,
            padding=dilation,
            stride=stride,
            bias=False,
        )
        self.conv2 = nn.Conv2d(
            planes,
            planes,
            kernel_size=3,
            dilation=dilation,
            padding=dilation,
            bias=False,
        )
        self.relu = nn.ReLU(inplace=True)

        self.norm1 = norm_layer(planes)
        self.norm2 = norm_layer(planes)
        if not stride == 1 or in_planes != planes:
            self.norm3 = norm_layer(planes)

        if stride == 1 and in_planes == planes:
            self.downsample = None
        else:
            self.downsample = nn.Sequential(
                nn.Conv2d(in_planes, planes, kernel_size=1, stride=stride), self.norm3
            )

    def forward(self, x):
        y = x
        y = self.relu(self.norm1(self.conv1(y)))
        y = self.relu(self.norm2(self.conv2(y)))

        if self.downsample is not None:
            x = self.downsample(x)

        return self.relu(x + y)


class CNNEncoder(nn.Module):
    def __init__(
        self,
        output_dim=128,
        norm_layer=nn.InstanceNorm2d,
        num_output_scales=1,
        **kwargs,
    ):
        super(CNNEncoder, self).__init__()
        self.num_branch = num_output_scales

        feature_dims = [64, 96, 128]

        self.conv1 = nn.Conv2d(
            3, feature_dims[0], kernel_size=7, stride=2, padding=3, bias=False
        )  # 1/2
        self.norm1 = norm_layer(feature_dims[0])
        self.relu1 = nn.ReLU(inplace=True)

        self.in_planes = feature_dims[0]
        self.layer1 = self._make_layer(
            feature_dims[0], stride=1, norm_layer=norm_layer
        )  # 1/2
        self.layer2 = self._make_layer(
            feature_dims[1], stride=2, norm_layer=norm_layer
        )  # 1/4

        # highest resolution 1/4 or 1/8
        stride = 2 if num_output_scales == 1 else 1
        self.layer3 = self._make_layer(
            feature_dims[2],
            stride=stride,
            norm_layer=norm_layer,
        )  # 1/4 or 1/8

        self.conv2 = nn.Conv2d(feature_dims[2], output_dim, 1, 1, 0)

        if self.num_branch > 1:
            if self.num_branch == 4:
                strides = (1, 2, 4, 8)
            elif self.num_branch == 3:
                strides = (1, 2, 4)
            elif self.num_branch == 2:
                strides = (1, 2)
            else:
                raise ValueError

            self.trident_conv = MultiScaleTridentConv(
                output_dim,
                output_dim,
                kernel_size=3,
                strides=strides,
                paddings=1,
                num_branch=self.num_branch,
            )

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, (nn.BatchNorm2d, nn.InstanceNorm2d, nn.GroupNorm)):
                if m.weight is not None:
                    nn.init.constant_(m.weight, 1)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

    def _make_layer(self, dim, stride=1, dilation=1, norm_layer=nn.InstanceNorm2d):
        layer1 = ResidualBlock(
            self.in_planes, dim, norm_layer=norm_layer, stride=stride, dilation=dilation
        )
        layer2 = ResidualBlock(
            dim, dim, norm_layer=norm_layer, stride=1, dilation=dilation
        )

        layers = (layer1, layer2)

        self.in_planes = dim
        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.norm1(x)
        x = self.relu1(x)

        x = self.layer1(x)  # 1/2
        x = self.layer2(x)  # 1/4
        x = self.layer3(x)  # 1/8 or 1/4

        x = self.conv2(x)

        if self.num_branch > 1:
            out = self.trident_conv([x] * self.num_branch)  # high to low res
        else:
            out = [x]

        return out
